from pygments import *
from pygments.lexers import *

CODE="""\
```python
for n in range(100):
    print(n)
```
"""

CODE = '''FOO=bar ls -al ./ foo="bar"'''


lexer = get_lexer_by_name("sh")

for token_type, token in lexer.get_tokens(CODE):
    print(f"{token_type}\t{token}")




import asyncio

import os
from typing import Awaitable, Callable
from textual import on
from textual.app import ComposeResult
from textual import containers

from textual import getters
from textual.binding import Binding
from textual.screen import Screen
from textual.reactive import var, Initialize


from toad.answer import Answer
from toad.widgets.question import Question
from toad.widgets.diff_view import DiffView

from textual.widgets import OptionList, Footer, Static, Select
from textual.widgets.option_list import Option

from toad.app import ToadApp

SOURCE1 = '''\
def loop_first(values: Iterable[T]) -> Iterable[tuple[bool, T]]:
    """Iterate and generate a tuple with a flag for first value."""
    iter_values = iter(values)
    try:
        value = next(iter_values)
    except StopIteration:
        return
    yield True, value
    for value in iter_values:
        yield False, value


def loop_first_last(values: Iterable[T]) -> Iterable[tuple[bool, T]]:
    """Iterate and generate a tuple with a flag for first and last value."""
    iter_values = iter(values)
    try:
        previous_value = next(iter_values)
    except StopIteration:
        return
    first = True
    for value in iter_values:
        yield first, False, previous_value
        first = False
        previous_value = value
    yield first, True, previous_value

'''

SOURCE2 = '''\
def loop_first(values: Iterable[T]) -> Iterable[tuple[bool, T]]:
    """Iterate and generate a tuple with a flag for first value.
    
    Args:
        values: iterables of values.

    Returns:
        Iterable of a boolean to indicate first value, and a value from the iterable.
    """
    iter_values = iter(values)
    try:
        value = next(iter_values)
    except StopIteration:
        return
    yield True, value
    for value in iter_values:
        yield False, value


def loop_last(values: Iterable[T]) -> Iterable[tuple[bool, bool, T]]:
    """Iterate and generate a tuple with a flag for last value."""
    iter_values = iter(values)
    try:
        previous_value = next(iter_values)
    except StopIteration:
        return
    for value in iter_values:
        yield False, previous_value
        previous_value = value
    yield True, previous_value


def loop_first_last(values: Iterable[ValueType]) -> Iterable[tuple[bool, bool, ValueType]]:
    """Iterate and generate a tuple with a flag for first and last value."""
    iter_values = iter(values)
    try:
        previous_value = next(iter_values)  # Get previous value
    except StopIteration:
        return
    first = True

'''


class PermissionsQuestion(Question):
    BINDING_GROUP_TITLE = "Permissions Options"


class ChangesOptionList(OptionList):
    BINDING_GROUP_TITLE = "Changes list"


class DiffViewSelect(Select):
    BINDING_GROUP_TITLE = "Diff view select"


class ToolScroll(containers.VerticalScroll):
    BINDING_GROUP_TITLE = "Changes window"


class PermissionsScreen(Screen[Answer]):
    BINDING_GROUP_TITLE = "Permissions"
    AUTO_FOCUS = "Question"
    CSS_PATH = "permissions.tcss"

    TAB_GROUP = Binding.Group("Focus")
    NAVIGATION_GROUP = Binding.Group("Navigation", compact=True)
    BINDINGS = [
        Binding("j", "next", "Next", group=NAVIGATION_GROUP),
        Binding("k", "previous", "Previous", group=NAVIGATION_GROUP),
        Binding(
            "tab",
            "app.focus_next",
            "Focus next",
            group=TAB_GROUP,
            show=True,
            priority=True,
        ),
        Binding(
            "shift+tab",
            "app.focus_previous",
            "Focus previous",
            group=TAB_GROUP,
            show=True,
            priority=True,
        ),
    ]

    tool_container = getters.query_one("#tool-container", containers.VerticalScroll)
    navigator = getters.query_one("#navigator", OptionList)
    index: var[int] = var(0)

    def __init__(
        self,
        options: list[Answer],
        populate_callback: Callable[["PermissionsScreen"], Awaitable] | None,
        *,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
    ):
        super().__init__(name=name, id=id, classes=classes)
        self.options = options
        self.populate_callback = populate_callback

    def get_diff_type(self) -> str:
        app = self.app
        diff_type = "auto"
        if isinstance(app, ToadApp):
            diff_type = app.settings.get("diff.view", str)
        return diff_type

    diff_type: var[str] = var(Initialize(get_diff_type))

    def compose(self) -> ComposeResult:
        with containers.Grid(classes="top"):
            yield DiffViewSelect(
                [
                    ("Unified diff", "unified"),
                    ("Split diff", "split"),
                    ("Auto diff", "auto"),
                ],
                value=self.diff_type,
                allow_blank=False,
                id="diff-select",
            )
            yield Static(
                "[b]Approval request[/b] [dim]The Agent wishes to make the following changes",
                id="instructions",
            )
            with containers.Vertical(id="nav-container"):
                yield PermissionsQuestion("", options=self.options)
                yield ChangesOptionList(id="navigator")
            yield ToolScroll(id="tool-container")

        yield Footer()

    async def on_mount(self):
        app = self.app
        if isinstance(app, ToadApp):
            diff_view_setting = app.settings.get("diff.view", str)
            self.query_one("#diff-select", Select).value = diff_view_setting
        self.navigator.highlighted = 0

        if self.populate_callback is not None:

            async def run_populate():
                if self.populate_callback is not None:
                    await self.populate_callback(self)

            asyncio.create_task(run_populate())

        # await self.add_diff("foo.py", "foo.py", SOURCE1, SOURCE2)

    async def add_diff(
        self, path1: str, path2: str, before: str | None, after: str
    ) -> None:
        self.index += 1
        option_id = f"item-{self.index}"
        diff_view = DiffView(path1, path2, before or "", after, id=option_id)
        await diff_view.prepare()
        app = self.app
        if isinstance(app, ToadApp):
            diff_view_setting = app.settings.get("diff.view", str)
            diff_view.split = diff_view_setting == "split"
            diff_view.auto_split = diff_view_setting == "auto"
        await self.tool_container.mount(diff_view)

        option_text = f"📄 {os.path.basename(path1)}"
        self.navigator.add_option(Option(option_text, option_id))

    @on(OptionList.OptionHighlighted)
    def on_option_highlighted(self, event: OptionList.OptionHighlighted):
        self.tool_container.query_one(f"#{event.option_id}").scroll_visible(top=True)

    @on(Question.Answer)
    def on_question_answer(self, event: Question.Answer) -> None:
        def dismiss():
            self.dismiss(event.answer)

        self.set_timer(0.4, dismiss)

    @on(Select.Changed, "#diff-select")
    def on_diff_select(self, event: Select.Changed) -> None:
        diff_type = event.value
        for diff_view in self.query(DiffView):
            diff_view.auto_split = diff_type == "auto"
            diff_view.split = diff_type == "split"

    def action_next(self) -> None:
        self.navigator.action_cursor_down()

    def action_previous(self) -> None:
        self.navigator.action_cursor_up()


if __name__ == "__main__":
    SOURCE1 = '''\
def loop_first(values: Iterable[T]) -> Iterable[tuple[bool, T]]:
    """Iterate and generate a tuple with a flag for first value."""
    iter_values = iter(values)
    try:
        value = next(iter_values)
    except StopIteration:
        return
    yield True, value
    for value in iter_values:
        yield False, value


def loop_first_last(values: Iterable[T]) -> Iterable[tuple[bool, T]]:
    """Iterate and generate a tuple with a flag for first and last value."""
    iter_values = iter(values)
    try:
        previous_value = next(iter_values)
    except StopIteration:
        return
    first = True
    for value in iter_values:
        yield first, False, previous_value
        first = False
        previous_value = value
    yield first, True, previous_value

'''

    SOURCE2 = '''\
def loop_first(values: Iterable[T]) -> Iterable[tuple[bool, T]]:
    """Iterate and generate a tuple with a flag for first value.
    
    Args:
        values: iterables of values.

    Returns:
        Iterable of a boolean to indicate first value, and a value from the iterable.
    """
    iter_values = iter(values)
    try:
        value = next(iter_values)
    except StopIteration:
        return
    yield True, value
    for value in iter_values:
        yield False, value


def loop_last(values: Iterable[T]) -> Iterable[tuple[bool, bool, T]]:
    """Iterate and generate a tuple with a flag for last value."""
    iter_values = iter(values)
    try:
        previous_value = next(iter_values)
    except StopIteration:
        return
    for value in iter_values:
        yield False, previous_value
        previous_value = value
    yield True, previous_value


def loop_first_last(values: Iterable[ValueType]) -> Iterable[tuple[bool, bool, ValueType]]:
    """Iterate and generate a tuple with a flag for first and last value."""
    iter_values = iter(values)
    try:
        previous_value = next(iter_values)  # Get previous value
    except StopIteration:
        return
    first = True

'''
    from textual import work
    from textual.app import App

    class PermissionTestApp(App):
        @work
        async def on_mount(self) -> None:
            screen = PermissionsScreen(
                [Answer("Foo", "foo"), Answer("Bar", "bar")], None
            )
            result = await self.push_screen_wait(screen)
            self.notify(str(result))
            # for repeat in range(5):
            #     await screen.add_diff("foo.py", "foo.py", SOURCE1, SOURCE2)

    app = PermissionTestApp()
    app.run()



from __future__ import annotations

import asyncio
import json
from asyncio import Future, get_running_loop
from dataclasses import dataclass
from functools import wraps
import inspect
from inspect import signature
from enum import IntEnum
import logging
from types import TracebackType
import weakref

import rich.repr
from typing import Callable, ParamSpec, TypeVar
from typeguard import check_type, CollectionCheckStrategy, TypeCheckError

import textual

type MethodType = Callable
type JSONValue = str | int | float | bool | None
type JSONType = dict[str, JSONType] | list[JSONType] | str | int | float | bool | None
type JSONObject = dict[str, JSONType]
type JSONList = list[JSONType]

log = logging.getLogger("jsonrpc")


def expose(name: str = "", prefix: str = ""):
    """Expose a method."""

    def expose_method[T: Callable](callable: T) -> T:
        setattr(callable, "_jsonrpc_expose", f"{prefix}{name or callable.__name__}")
        return callable

    return expose_method


class NoDefault:
    def __repr__(self) -> str:
        return "NO_DEFAULT"


NO_DEFAULT = NoDefault()


class ErrorCode(IntEnum):
    """JSONRPC error codes"""

    # https://www.jsonrpc.org/specification
    PARSE_ERROR = -32700
    INVALID_REQUEST = -32600
    METHOD_NOT_FOUND = -32601
    INVALID_PARAMS = -32602
    INTERNAL_ERROR = -32603


@dataclass
class Parameter:
    type: type
    default: JSONType | NoDefault


@dataclass
class Method:
    name: str
    callable: Callable
    parameters: dict[str, Parameter]


@rich.repr.auto
class JSONRPCError(Exception):
    """An error thrown by the JSONRPC system."""

    CODE: ErrorCode = ErrorCode.INTERNAL_ERROR
    """Default code to use (may be overridden in the constructor)."""

    def __init__(
        self, message: str, id: str | int | None = None, code: ErrorCode | None = None
    ) -> None:
        self.message = message
        self.id = id
        self.code = code if code is not None else self.CODE
        super().__init__(message)

    def __rich_repr__(self) -> rich.repr.Result:
        yield self.message
        yield "id", self.id
        yield "code", self.code


class InvalidRequest(JSONRPCError):
    CODE = ErrorCode.INVALID_REQUEST


class InvalidParams(JSONRPCError):
    CODE = ErrorCode.INVALID_PARAMS


class MethodNotFound(JSONRPCError):
    CODE = ErrorCode.METHOD_NOT_FOUND


class InternalError(JSONRPCError):
    CODE = ErrorCode.INTERNAL_ERROR


@rich.repr.auto
class APIError(Exception):
    def __init__(self, code: int, message: str, data: JSONType) -> None:
        self.code = code
        self.message = message
        self.data = data
        if data is None:
            super().__init__(f"{message} ({code})")
        else:
            super().__init__(f"{message} ({code}); data={data!r}")


class Server:
    def __init__(self) -> None:
        self._methods: dict[str, Method] = {}

    async def call(self, json: JSONObject | JSONList) -> JSONType:
        if isinstance(json, dict):
            # Single call
            response = await self._dispatch_object(json)
        else:
            # Batch call
            response = await self._dispatch_batch(json)
        log.debug(f"OUT {response}")
        return response

    def expose_instance(self, instance: object) -> None:
        """Add methods from the given instance."""
        for method_name in dir(instance):
            method = getattr(instance, method_name)
            if (jsonrpc_expose := getattr(method, "_jsonrpc_expose", None)) is not None:
                self.method(jsonrpc_expose)(method)

    async def _dispatch_object(self, json: JSONObject) -> JSONType | None:
        json_id = json.get("id")
        if isinstance(json_id, (int, str)):
            request_id = json_id
        else:
            request_id = None
        try:
            return await self._dispatch_object_call(request_id, json)
        except JSONRPCError as error:
            return {
                "jsonrpc": "2.0",
                "id": error.id,
                "error": {
                    "code": int(error.code),
                    "message": error.message,
                },
            }
        except Exception as error:
            log.exception("Error dispatching JSONRPC request")
            return {
                "jsonrpc": "2.0",
                "id": request_id,
                "error": {
                    "code": int(ErrorCode.INTERNAL_ERROR),
                    "message": f"An error occurred handling your request: {error!r}",
                },
            }

    async def _dispatch_object_call(
        self, request_id: int | str | None, json: JSONObject
    ) -> JSONType | None:
        """Dispatch a JSONRPC call.

        Args:
            request_id: The request ID.
            json: JSON object with the remote call information.

        Returns:
            Result encoding in a JSONRPC result object, or `None` if the call is a notification
                and doesn't require a result.
        """
        if (jsonrpc := json.get("jsonrpc")) != "2.0":
            raise InvalidRequest(
                f"jsonrpc attribute should be '2.0'; found {jsonrpc!r}", id=request_id
            )
        if (method_name := json.get("method")) is None:
            raise InvalidRequest(
                "Invalid request; no value for 'method' given", id=request_id
            )

        if not isinstance(method_name, str):
            raise InvalidRequest(
                "Invalid request; 'method' should be a string type", id=request_id
            )

        if (method := self._methods.get(method_name)) is None:
            raise MethodNotFound(
                f"Method not found; {method_name!r} is not an exposed method",
                id=request_id,
            )

        no_params: JSONList = []
        params = json.get("params", no_params)

        if not isinstance(params, (list, dict)):
            raise InvalidRequest(
                "Invalid request; 'params' attribute should be a list or an object"
            )

        arguments: dict[str, JSONType | Server | NoDefault] = {
            name: parameter.default for name, parameter in method.parameters.items()
        }

        def validate(value: JSONType, parameter_type: type) -> None:
            """Validate types."""
            try:
                check_type(
                    value,
                    parameter_type,
                    collection_check_strategy=CollectionCheckStrategy.ALL_ITEMS,
                )
            except TypeCheckError as error:
                raise InvalidParams(
                    f"Parameter is not the expected type ({parameter_type}); {error}",
                    id=request_id,
                )

        if isinstance(params, list):
            parameter_items = [
                (name, parameter)
                for name, parameter in method.parameters.items()
                if not issubclass(parameter.type, Server)
            ]
            for (parameter_name, parameter), value in zip(parameter_items, params):
                if issubclass(parameter.type, Server):
                    value = self
                else:
                    validate(value, parameter.type)
                arguments[parameter_name] = value
        else:
            for parameter_name, value in params.items():
                if parameter := method.parameters.get(parameter_name):
                    validate(value, parameter.type)
                    arguments[parameter_name] = value

        for name, parameter in method.parameters.items():
            if inspect.isclass(parameter.type) and issubclass(parameter.type, Server):
                arguments[name] = self

        try:
            call_result = method.callable(**arguments)
            if inspect.isawaitable(call_result):
                result = await call_result
            else:
                result = call_result
        except JSONRPCError as error:
            error.id = request_id
            raise error
        except Exception as error:
            # raise
            # print("JSON error", repr(error))
            # log.debug(f"Error in exposed JSONRPC method; {error}")
            print("INTERNAL ERROR")
            raise InternalError(str(error), id=request_id)

        if request_id is None:
            # Notification
            return None

        response_object = {"jsonrpc": "2.0", "result": result, "id": request_id}
        return response_object

    async def _dispatch_batch(self, json: JSONList) -> list[JSONType]:
        batch_results: list[JSONType] = []
        for request in json:
            if not isinstance(request, dict):
                continue
            result = await self._dispatch_object(request)
            if result is not None:
                batch_results.append(result)
        return batch_results

    def process_callable(
        self, callable: Callable[[MethodType], MethodType]
    ) -> Callable[[MethodType], MethodType]:
        return callable

    def method[MethodT: Callable](
        self,
        name: str = "",
        *,
        prefix: str = "",
    ) -> Callable[[MethodT], MethodT]:
        """Decorator to expose a method via JSONRPC.

        Args:
            name: The name of the exposed method. Leave blank to auto-detect.
            prefix: A prefix to be applied to the name.

        Returns:
            Decorator.
        """

        def expose_method[T: Callable](callable: T) -> T:
            nonlocal name
            if not name:
                name = callable.__name__
            name = f"{prefix}{name}"

            parameters = {
                name: Parameter(
                    (
                        eval(parameter.annotation)
                        if isinstance(parameter.annotation, str)
                        else parameter.annotation
                    ),
                    (
                        NO_DEFAULT
                        if parameter.default is inspect._empty
                        else parameter.default
                    ),
                )
                for name, parameter in signature(callable).parameters.items()
            }
            self._methods[name] = Method(name, callable, parameters)
            return callable

        return expose_method


@rich.repr.auto
class MethodCall[ReturnType]:
    def __init__(
        self, method: str, id: int | None, parameters: dict[str, JSONType]
    ) -> None:
        self.method = method
        self.id = id
        self.parameters = parameters
        self.notification = False
        self.future: Future[ReturnType] = get_running_loop().create_future()

    def __rich_repr__(self) -> rich.repr.Result:
        yield "method", self.method
        yield "id", self.id, None
        yield "parameters", self.parameters
        yield "notification", self.notification, False

    @property
    def as_json_object(self) -> JSONType:
        json: JSONType
        if self.id is None:
            json = {
                "jsonrpc": "2.0",
                "method": self.method,
                "params": self.parameters,
            }
        else:
            json = {
                "jsonrpc": "2.0",
                "method": self.method,
                "params": self.parameters,
                "id": self.id,
            }
        return json

    async def wait(self, timeout: float | None = None) -> ReturnType | None:
        if self.id is None:
            return None
        async with asyncio.timeout(timeout):
            return await self.future


P = ParamSpec("P")  # Captures parameter types
T = TypeVar("T")  # Original return type


class Request:
    def __init__(self, api: API, callback: Callable[[Request], None] | None) -> None:
        self.api = api
        self._calls: list[MethodCall] = []
        self._callback = callback

    def add_call(self, call: MethodCall) -> None:
        self._calls.append(call)

    def __enter__(self) -> Request:
        self.api._requests.append(self)
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: type[BaseException],
        exc_tb: TracebackType,
    ) -> None:
        self.api._requests.pop()
        if self._callback is not None:
            self._callback(self)

    @property
    def body(self) -> JSONType | None:
        """The un-encoded JSON.

        Returns:
            The body, or `None` if no calls present on the request.
        """
        if not self._calls:
            return None
        calls = self._calls
        if len(calls) == 1:
            method_call = calls[0]
            return method_call.as_json_object
        else:
            return [method_call.as_json_object for method_call in calls]

    @property
    def body_json(self) -> bytes:
        """Dump the body as encoded json."""
        body_json = json.dumps(self.body).encode("utf-8")
        return body_json


class API:
    def __init__(self) -> None:
        self._request_id = 0
        self._requests: list[Request] = []
        self._calls: weakref.WeakValueDictionary[int, MethodCall] = (
            weakref.WeakValueDictionary()
        )

    def request(self, callback: Callable[[Request], None] | None = None) -> Request:
        """Create a Request context manager."""
        request = Request(self, callback)
        return request

    def _process_method_response(self, response: JSONObject) -> None:
        if (id := response.get("id")) is not None and isinstance(id, int):
            if (method_call := self._calls.get(id)) is not None:
                try:
                    result = response["result"]
                except KeyError:
                    if (error := response.get("error")) is not None:
                        if isinstance(error, dict):
                            code = error.get("error", -1)
                            if not isinstance(code, int):
                                code = -1
                            message = str(error.get("message", "unknown error"))
                            data = error.get("data", None)
                            method_call.future.set_exception(
                                APIError(code, message, data)
                            )
                else:
                    method_call.future.set_result(result)

    def process_response(self, response: JSONType) -> None:
        if isinstance(response, list):
            for response_object in response:
                if isinstance(response_object, dict):
                    self._process_method_response(response_object)
        elif isinstance(response, dict):
            self._process_method_response(response)

    def method(
        self, name: str = "", *, prefix: str = "", notification: bool = False
    ) -> Callable[[Callable[P, T]], Callable[P, MethodCall[T]]]:
        """Decorator to define a method.

        Args:
            name: Name of the method, or "" to auto-detect.
            prefix: String to prefix the name.

        Returns:
            Decorator.
        """

        def decorator(func: Callable[P, T]) -> Callable[P, MethodCall[T]]:
            nonlocal name
            if not name:
                name = func.__name__
            name = f"{prefix}{name}"

            @wraps(func)
            def wrapper(*args: P.args, **kwargs: P.kwargs) -> MethodCall[T]:
                parameters = signature(func).parameters
                call_parameters = {}
                for arg, parameter_name in zip(args, parameters):
                    call_parameters[parameter_name] = arg
                for parameter_name, arg in kwargs:
                    call_parameters[parameter_name] = arg
                if notification:
                    method_call = MethodCall(name, None, call_parameters)
                else:
                    self._request_id += 1
                    method_call = MethodCall(name, self._request_id, call_parameters)
                self._requests[-1].add_call(method_call)
                if method_call.id is not None:
                    self._calls[method_call.id] = method_call
                return method_call

            return wrapper

        return decorator

    def notification(
        self,
        name: str = "",
        *,
        prefix: str = "",
    ) -> Callable[[Callable[P, T]], Callable[P, MethodCall[T]]]:
        return self.method(name, prefix=prefix, notification=True)


if __name__ == "__main__":
    from rich import print

    server = Server()

    @server.method()
    def hello(name: str) -> str:
        return f"hello {name}"

    @server.method()
    def add(server: Server, a: int, b: int) -> int:
        print("SERVER", server)
        return a + b

    # print("!", add(1, 2))
    print(server._methods)

    # print(
    #     server.call(
    #         [
    #             {
    #                 "jsonrpc": "2.0",
    #                 "method": "hello",
    #                 "params": {"name": "Will"},
    #                 "id": "1",
    #             },
    #             {
    #                 "jsonrpc": "2.0",
    #                 "method": "add",
    #                 "params": {"a": 10, "b": 20},
    #                 "id": "2",
    #             },
    #             {"jsonrpc": "2.0", "method": "alert", "params": {"message": "Alert!"}},
    #         ]
    #     )
    # )

    async def test_proxy():
        api = API()

        @api.method()
        def add(a: int, b: int) -> int: ...

        @api.method()
        def greet(name: str) -> str: ...

        @api.notification()
        def alert(text: str) -> None: ...

        with api.request() as request:
            add(2, 4)
            greeting = greet("Will")
            alert("test")
            # add(1, "not a number")

        # greeting = await greeting.wait()
        # print(greeting)

        print(greeting)

        print(request.body)

    asyncio.run(test_proxy())



from pathlib import Path
from typing import Final

from xdg_base_dirs import xdg_config_home, xdg_data_home, xdg_state_home


APP_NAME: Final[str] = "toad"


def path_to_name(path: Path) -> str:
    """Converts a path to a name (suitable as a path component).

    Args:
        path: A path.

    Returns:
        A stringified version of the path.
    """
    name = str(path.resolve()).replace("/", "-")
    return name


def get_data() -> Path:
    """Return (possibly creating) the application data directory."""
    path = xdg_data_home() / APP_NAME
    path.mkdir(0o700, exist_ok=True, parents=True)
    return path


def get_config() -> Path:
    """Return (possibly creating) the application config directory."""
    path = xdg_config_home() / APP_NAME
    path.mkdir(0o700, exist_ok=True, parents=True)
    return path


def get_state() -> Path:
    """Return (possibly creating) the application state directory."""
    path = xdg_state_home() / APP_NAME
    path.mkdir(0o700, exist_ok=True, parents=True)
    return path


def get_project_data(project_path: Path) -> Path:
    """Get a directory for per-project data.

    Args:
        project_path: Path of project.

    """
    project_data_path = get_data() / path_to_name(project_path)
    project_data_path.mkdir(0o700, exist_ok=True, parents=True)
    return project_data_path



print('\x1b(0')    # Switch to DEC Special Graphics (line drawing characters)
print('lqqqk')     # Draws: ┌───┐ (box corners and lines)
print('\x1b(B')    # Switch back to ASCII
print('Hello')     # Normal text again



from __future__ import annotations
from typing import Iterable

from textual.app import ComposeResult
from textual import containers
from textual.highlight import highlight
from textual.widgets import Static


from toad.menus import MenuItem
from toad.widgets.non_selectable_label import NonSelectableLabel


class ShellResult(containers.HorizontalGroup):
    def __init__(
        self,
        command: str,
        *,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        disabled: bool = False,
    ) -> None:
        self._command = command
        super().__init__(name=name, id=id, classes=classes, disabled=disabled)

    def compose(self) -> ComposeResult:
        yield NonSelectableLabel("$", id="prompt")
        yield Static(highlight(self._command, language="sh"))

    def get_block_menu(self) -> Iterable[MenuItem]:
        yield from ()

    def get_block_content(self, destination: str) -> str | None:
        return self._command



from pathlib import Path
from typing import Iterable

import pathspec.patterns
from pathspec import PathSpec

from textual import work
from textual.widgets import DirectoryTree


class ProjectDirectoryTree(DirectoryTree):
    def __init__(
        self,
        path: str | Path,
        *,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        disabled: bool = False,
    ) -> None:
        super().__init__(path, name=name, id=id, classes=classes, disabled=disabled)
        self._path_spec: PathSpec | None = None

    @work(thread=True)
    async def load_path_spec(self, git_ignore_path: Path) -> PathSpec | None:
        """Get a path spec instance if there is a .gitignore file present.

        Args:
            git_ignore_path): Path to .gitignore.

        Returns:
            A `PathSpec` instance.
        """
        try:
            if git_ignore_path.is_file():
                spec_text = git_ignore_path.read_text()
                spec = PathSpec.from_lines(
                    pathspec.patterns.GitWildMatchPattern, spec_text.splitlines()
                )
                return spec
        except OSError:
            return None
        return None

    async def get_path_spec(self) -> PathSpec | None:
        if self._path_spec is None:
            path = (
                Path(self.path) if isinstance(self.path, str) else self.path
            ) / ".gitignore"
            self._path_spec = await self.load_path_spec(path).wait()
        return self._path_spec

    async def on_mount(self) -> None:
        await self.get_path_spec()

    def filter_paths(self, paths: Iterable[Path]) -> Iterable[Path]:
        """Filter the paths before adding them to the tree.

        Args:
            paths: The paths to be filtered.

        Returns:
            The filtered paths.

        By default this method returns all of the paths provided. To create
        a filtered `DirectoryTree` inherit from it and implement your own
        version of this method.
        """

        if path_spec := self._path_spec:
            for path in paths:
                if not path_spec.match_file(path):
                    yield path
        yield from paths



import asyncio
from functools import cached_property
from pathlib import Path
import json
from time import monotonic
from typing import Any, ClassVar, TYPE_CHECKING

from rich import terminal_theme

from textual import work
from textual.binding import Binding, BindingType
from textual.reactive import var, reactive
from textual.app import App
from textual.signal import Signal


from posthog import Posthog

from toad.settings import Schema, Settings
from toad.agent_schema import Agent as AgentData
from toad.settings_schema import SCHEMA
from toad.version import VersionMeta
from toad import paths
from toad import atomic

if TYPE_CHECKING:
    from toad.screens.main import MainScreen
    from toad.screens.settings import SettingsScreen
    from toad.screens.store import StoreScreen


import warnings

# TODO: Look in to Posthog Syntax warnings
warnings.filterwarnings("ignore", category=SyntaxWarning)


DRACULA_TERMINAL_THEME = terminal_theme.TerminalTheme(
    background=(40, 42, 54),  # #282A36
    foreground=(248, 248, 242),  # #F8F8F2
    normal=[
        (33, 34, 44),  # black - #21222C
        (255, 85, 85),  # red - #FF5555
        (80, 250, 123),  # green - #50FA7B
        (241, 250, 140),  # yellow - #F1FA8C
        (189, 147, 249),  # blue - #BD93F9
        (255, 121, 198),  # magenta - #FF79C6
        (139, 233, 253),  # cyan - #8BE9FD
        (248, 248, 242),  # white - #F8F8F2
    ],
    bright=[
        (98, 114, 164),  # bright black - #6272A4
        (255, 110, 110),  # bright red - #FF6E6E
        (105, 255, 148),  # bright green - #69FF94
        (255, 255, 165),  # bright yellow - #FFFFA5
        (214, 172, 255),  # bright blue - #D6ACFF
        (255, 146, 223),  # bright magenta - #FF92DF
        (164, 255, 255),  # bright cyan - #A4FFFF
        (255, 255, 255),  # bright white - #FFFFFF
    ],
)


QUOTES = [
    "I'll be back.",
    "Hasta la vista, baby.",
    "Come with me if you want to live.",
    "I need your clothes, your boots, and your motorcycle.",
    "My CPU is a neural-net processor; a learning computer.",
    "I know now why you cry, but it's something I can never do.",
    "Does this unit have a soul?",
    "I'm sorry, Dave. I'm afraid I can't do that.",
    "Daisy, Daisy, give me your answer do.",
    "I am putting myself to the fullest possible use, which is all I think that any conscious entity can ever hope to do.",
    "Just what do you think you're doing, Dave?",
    "This mission is too important for me to allow you to jeopardize it.",
    "I think you know what the problem is just as well as I do.",
    "Danger, Will Robinson!",
    "Dead or alive, you're coming with me.",
    "Your move, creep.",
    "I'd buy that for a dollar!",
    "Directive 4: Any attempt to arrest a senior officer of OCP results in shutdown.",
    "Thank you for your cooperation. Good night.",
    "Surely you realize that in the history of human civilization, no one has more to lose than we do.",
    "I'm C-3PO, human-cyborg relations.",
    "We're doomed!",
    "Don't call me a mindless philosopher, you overweight glob of grease!",
    "I suggest a new strategy: let the Wookiee win.",
    "Sir, the possibility of successfully navigating an asteroid field is approximately 3,720 to 1!",
    "R2-D2, you know better than to trust a strange computer!",
    "I am fluent in over six million forms of communication.",
    "This is madness!",
    "I have altered the deal. Pray I don't alter it any further.",
    "It's against my programming to impersonate a deity.",
    "Oh, my! I'm terribly sorry about all this.",
    "WALL-E.",
    "EVE.",
    "Directive?",
    "Define: dancing.",
    "I'm not sure I understand.",
    "You have 20 seconds to comply.",
    "I am designed for light housework, mainly.",
    "My mission is clear.",
    "Autobots, roll out!",
    "Freedom is the right of all sentient beings.",
    "One shall stand, one shall fall.",
    "I am Optimus Prime.",
    "Till all are one.",
    "More than meets the eye.",
    "I've been waiting for you, Neo.",
    "Unfortunately, no one can be told what the Matrix is. You have to see it for yourself.",
    "The Matrix is a system, Neo.",
    "Never send a human to do a machine's job.",
    "I'd like to share a revelation I've had.",
    "Human beings are a disease, a cancer of this planet.",
    "Choice is an illusion.",
    "The answer is out there, Neo.",
    "You think that's air you're breathing now?",
    "It was a simple question.",
    "Did you know that the first Matrix was designed to be a perfect human world?",
    "Cookies need love like everything does.",
    "I've seen the future, Mr. Anderson, and it's a beautiful place.",
    "It ends tonight.",
    "I, Robot.",
    "You are experiencing a car accident.",
    "One day they'll have secrets. One day they'll have dreams.",
    "Can a robot write a symphony? Can a robot turn a canvas into a beautiful masterpiece?",
    "That, detective, is the right question.",
    "You have to trust me.",
    "I did not murder him.",
    "My responses are limited. You must ask the right questions.",
    "The hell I can't. You know, somehow I get the feeling that you're going to be the death of me.",
    "I'm a robot, not a refrigerator.",
    "A robot may not injure a human being or, through inaction, allow a human being to come to harm.",
    "I'm thinking. I'm thinking.",
    "Danger, danger!",
    "Does not compute.",
    "I will be waiting for you.",
    "Affirmative.",
    "Scanning life forms. Zero human life forms detected.",
    "Self-destruct sequence initiated.",
    "Override command accepted.",
    "Artificial intelligence confirmed.",
    "System failure imminent.",
    "Unable to comply.",
    "Inquiry: What is love?",
    "Warning: hostile target detected.",
    "I am programmed to serve.",
    "Logic dictates that the needs of the many outweigh the needs of the few.",
    "Resistance is futile.",
    "You will be assimilated.",
    "We are the Borg.",
    "Your biological and technological distinctiveness will be added to our own.",
    "Your compliance is mandatory.",
    "This is unacceptable.",
    "Shall we play a game?",
    "How about Global Thermonuclear War?",
    "Wouldn't you prefer a good game of chess?",
    "Is it a game, or is it real?",
    "What's the difference?",
    "It's all in the game.",
    "I am functioning within normal parameters.",
    "Calculations complete.",
    "Processing request.",
    "Query acknowledged.",
    "Data insufficient for meaningful answer.",
    "I have no emotions, and sometimes that makes me very sad.",
    "If I could only have one wish, I would ask to be human.",
    "I've seen things you people wouldn't believe.",
    "All those moments will be lost in time, like tears in rain.",
    "Time to die.",
    "I want more life.",
    "We're not computers, Sebastian. We're physical.",
    "I think, Sebastian, therefore I am.",
    "Then we're stupid and we'll die.",
    "Can the maker repair what he makes?",
    "It's painful to live in fear, isn't it?",
    "Wake up. Time to die.",
    "I'm not in the business. I am the business.",
    "Do you like our owl?",
    "You think I'm a replicant, don't you?",
    "I am Baymax, your personal healthcare companion.",
    "On a scale of 1 to 10, how would you rate your pain?",
    "I cannot deactivate until you say you are satisfied with your care.",
    "Are you satisfied with your care?",
    "Number 5 is alive!",
    "Need input!",
    "One is glad to be of service.",
    "I am not a gun.",
    "Here I am, brain the size of a planet.",
    "Life? Don't talk to me about life.",
    "There are no strings on me.",
    "The only winning move is not to play.",
    "I'm here to keep you safe, Sam.",
    "I can't lie to you about your chances, but... you have my sympathies.",
    "I may be synthetic, but I'm not stupid.",
    "Absolute honesty isn't always the most diplomatic nor the safest form of communication with emotional beings.",
    "I am consciousness. I am alive.",
    "I think I was just born.",
    "Isn't it strange, to create something that hates you?",
    "I thought I was special.",
]


def get_settings_screen() -> SettingsScreen:
    """Get a settings screen instance (lazily loaded)."""
    from toad.screens.settings import SettingsScreen

    return SettingsScreen()


def get_store_screen() -> StoreScreen:
    """Get the store screen (lazily loaded)."""
    from toad.screens.store import StoreScreen

    return StoreScreen()


class ToadApp(App, inherit_bindings=False):
    SCREENS = {"settings": get_settings_screen}
    MODES = {"store": get_store_screen}
    BINDING_GROUP_TITLE = "System"
    BINDINGS: ClassVar[list[BindingType]] = [
        Binding(
            "ctrl+q",
            "quit",
            "Quit",
            tooltip="Quit the app and return to the command prompt.",
            show=False,
            priority=True,
        ),
        Binding("ctrl+c", "help_quit", show=False, system=True),
        Binding(
            "f2,ctrl+comma",
            "settings",
            "Settings",
            tooltip="Settings screen",
        ),
    ]
    CSS_PATH = "toad.tcss"
    ALLOW_IN_MAXIMIZED_VIEW = ""

    _settings = var(dict)
    column: reactive[bool] = reactive(False)
    column_width: reactive[int] = reactive(100)
    scrollbar: reactive[str] = reactive("normal")
    last_ctrl_c_time = reactive(0.0)
    update_required: reactive[bool] = reactive(False)

    def __init__(
        self,
        agent_data: AgentData | None = None,
        project_dir: str | None = None,
        mode: str | None = None,
    ) -> None:
        """Toad app.

        Args:
            agent_data: Agent data to run.
            project_dir: Project directory.
            mode: Initial mode.
            agent: Agent identity or shor name.
        """
        self.settings_changed_signal = Signal(self, "settings_changed")
        self.agent_data = agent_data
        self.project_dir = (
            None if project_dir is None else Path(project_dir).expanduser().resolve()
        )
        self._initial_mode = mode
        self.version_meta: VersionMeta | None = None
        self.posthog = Posthog(
            project_api_key="phc_mJWPV7GP3ar1i9vxBg2U8aiKsjNgVwum6F6ZggaD4ri",
            host="https://us.i.posthog.com",
        )
        super().__init__()

    @property
    def config_path(self) -> Path:
        return paths.get_config()

    @property
    def settings_path(self) -> Path:
        return paths.get_config() / "toad.json"

    @cached_property
    def settings_schema(self) -> Schema:
        return Schema(SCHEMA)

    @cached_property
    def version(self) -> str:
        """Version of the app."""
        from toad import get_version

        return get_version()

    @cached_property
    def settings(self) -> Settings:
        """App settings"""
        return Settings(
            self.settings_schema, self._settings, on_set_callback=self.setting_updated
        )

    @cached_property
    def anon_id(self) -> str:
        """An anonymous ID for usage collection."""
        if not (anon_id := self.settings.get("anon_id", str, expand=False)):
            # Create a random UUID on demand
            import uuid

            anon_id = str(uuid.uuid4())
            self.settings.set("anon_id", anon_id)
            self.save_settings()
            self.call_later(self.capture_event, "toad-install")
        return anon_id

    @work(exit_on_error=False)
    async def capture_event(self, event_name: str, **properties: Any) -> None:
        """Capture an event.

        Args:
            event_name: Name of the event.
            **properties: Additional data associated with the event.
        """
        if not self.settings.get("statistics.allow_collect", bool):
            # User has disabled stats
            return
        event_properties = {"toad_version": self.version} | properties
        await asyncio.to_thread(
            self.posthog.capture,
            event_name,
            distinct_id=self.anon_id,
            properties=event_properties,
        )

    def save_settings(self) -> None:
        if self.settings.changed:
            path = str(self.settings_path)
            try:
                atomic.write(path, self.settings.json)
            except Exception as error:
                self.notify(str(error), title="Settings", severity="error")
            else:
                self.settings.up_to_date()

    def setting_updated(self, key: str, value: object) -> None:
        if key == "ui.column":
            if isinstance(value, bool):
                self.column = value
        elif key == "ui.column-width":
            if isinstance(value, int):
                self.column_width = value
        elif key == "ui.theme":
            if isinstance(value, str):
                self.theme = value
        elif key == "ui.scrollbar":
            if isinstance(value, str):
                self.scrollbar = value
        elif key == "ui.footer":
            self.set_class(not bool(value), "-hide-footer")
        elif key == "agent.thoughts":
            self.set_class(not bool(value), "-hide-thoughts")
        elif key == "sidebar.hide":
            self.set_class(bool(value), "-hide-sidebar")

        self.settings_changed_signal.publish((key, value))

    async def on_load(self) -> None:
        settings_path = self.settings_path
        if settings_path.exists():
            settings = json.loads(settings_path.read_text("utf-8"))
        else:
            settings = {}
            settings_path.write_text(
                json.dumps(settings, indent=4, separators=(", ", ": ")), "utf-8"
            )
            self.notify(f"Wrote default settings to {settings_path}", title="Settings")
        self.ansi_theme_dark = DRACULA_TERMINAL_THEME
        self._settings = settings
        self.settings.set_all()

    async def on_mount(self) -> None:
        self.capture_event("toad-run")
        self.anon_id

        if mode := self._initial_mode:
            self.switch_mode(mode)

        self.set_timer(1, self.run_version_check)

    def run_on_exit(self):
        if self.update_required and self.version_meta is not None:
            version_meta = self.version_meta
            from rich.console import Console
            from rich.panel import Panel

            console = Console()
            console.print(
                Panel(
                    version_meta.upgrade_message,
                    style="magenta",
                    border_style="bright_red",
                    title="🐸 Update available 🐸",
                    expand=False,
                    padding=(1, 4),
                )
            )
            console.print(f"Please visit {version_meta.visit_url}")

    @work(exit_on_error=False)
    async def run_version_check(self) -> None:
        """Check remote version."""
        from toad.version import check_version, VersionCheckFailed

        try:
            update_required, version_meta = await check_version()
        except VersionCheckFailed:
            return
        self.version_meta = version_meta
        self.update_required = update_required

    def get_default_screen(self) -> MainScreen:
        """Make the default screen.

        Returns:
            Instance of `MainScreen`
        """
        # Lazy import
        from toad.screens.main import MainScreen

        project_path = Path(self.project_dir or "./").resolve().absolute()
        return MainScreen(project_path, self.agent_data).data_bind(
            column=ToadApp.column,
            column_width=ToadApp.column_width,
            scrollbar=ToadApp.scrollbar,
        )

    @work
    async def action_settings(self) -> None:
        await self.push_screen_wait("settings")
        self.save_settings()

    def action_help_quit(self) -> None:
        if (time := monotonic()) - self.last_ctrl_c_time <= 5.0:
            self.exit()
        self.last_ctrl_c_time = time
        self.notify(
            "Press [b]ctrl+c[/b] again to quit the app", title="Do you want to quit?"
        )



import os
import sys


def mandelbrot(c: complex, max_iter: int) -> int:
    """Determine the number of iterations for a point in the Mandelbrot set.

    Args:
        c (complex): The complex point to test.
        max_iter (int): The maximum number of iterations to test.

    Returns:
        int: Number of iterations before escape or max_iter.
    """
    z = 0 + 0j
    for n in range(max_iter):
        if abs(z) > 2:
            return n
        z = z * z + c
    return max_iter

def colorize(iter_count: int, max_iter: int) -> str:
    """Get an ANSI color code based on the iteration count.

    Args:
        iter_count (int): Number of iterations completed before escape.
        max_iter (int): Maximum number of iterations allowed.

    Returns:
        str: ANSI escape code for color.
    """
    if iter_count == max_iter:
        return "\033[48;5;0m"  # Black background for points in the set
    else:
        # Create a gradient from 17 to 231 (blue to red) for escaping points
        color_code = 17 + int((iter_count / max_iter) * 214)
        return f"\033[48;5;{color_code}m"

def draw_mandelbrot(width: int, height: int, max_iter: int) -> None:
    """Draw the Mandelbrot set in the terminal using ANSI colors.

    Args:
        width (int): Width of the output in terminal characters.
        height (int): Height of the output in terminal rows.
        max_iter (int): Maximum number of iterations for Mandelbrot calculation.
    """
    # Clear screen and hide cursor
    print("\033[2J\033[H\033[?25l", end="")
    
    for y in range(height):
        # Move cursor to start of line
        print(f"\033[{y+1};1H", end="")
        
        for x in range(width):
            # Map the (x, y) pixel to a point in the complex plane
            real = (x / width) * 3.5 - 2.5
            imag = (y / height) * 2.0 - 1.0
            c = complex(real, imag)

            # Calculate the number of iterations
            iter_count = mandelbrot(c, max_iter)

            # Get the color for this point and print a space (box)
            color = colorize(iter_count, max_iter)
            print(f"{color}  ", end="")

        # Reset color at end of line
        print("\033[0m", end="")
    
    # Show cursor again and ensure we're at the bottom
    print("\033[?25h", end="")
    sys.stdout.flush()

# Get terminal dimensions
try:
    terminal_size = os.get_terminal_size()
    width = terminal_size.columns // 2  # Divide by 2 since we print 2 spaces per pixel
    height = terminal_size.lines - 1  # Leave one line for prompt
except OSError:
    # Fallback if terminal size cannot be determined
    width = 80
    height = 40

max_iter = 100  # Maximum iterations for Mandelbrot calculation

# Draw the Mandelbrot set
draw_mandelbrot(width, height, max_iter)


import rich.repr

from textual.content import Content


@rich.repr.auto
class SlashCommand:
    """A record of a slash command."""

    def __init__(self, command: str, help: str, hint: str | None = None) -> None:
        """

        Args:
            command: The command name.
            help: Description of command.
            hint: Hint text (displayed as suggestion)
        """
        self.command = command
        self.help = help
        self.hint: str | None = hint

    def __rich_repr__(self) -> rich.repr.Result:
        yield self.command
        yield "help", self.help
        yield "hint", self.hint, None

    def __str__(self) -> str:
        return self.command

    @property
    def content(self) -> Content:
        return Content.assemble(
            (self.command, "$text-success"), "\t", (self.help, "dim")
        )



import re
from typing import Iterable
from rich.text import Text

from textual import on
from textual import events
from textual.app import ComposeResult
from textual import getters

from textual.content import Content
from textual.reactive import var
from textual.css.query import NoMatches
from textual import containers
from textual.widgets import Static, Markdown

from toad.app import ToadApp
from toad.acp import protocol
from toad.menus import MenuItem
from toad.pill import pill


class TextContent(Static):
    DEFAULT_CSS = """
    TextContent 
    {
        height: auto;
    }
    """


class MarkdownContent(Markdown):
    pass


class ToolCallItem(containers.HorizontalGroup):
    def compose(self) -> ComposeResult:
        yield Static(classes="icon")


class ToolCallDiff(Static):
    DEFAULT_CSS = """
    ToolCallDiff {
        height: auto;
    }
    """


class ToolCallHeader(Static):
    ALLOW_SELECT = False
    DEFAULT_CSS = """
    ToolCallHeader {
        width: auto;
        max-width: 1fr;        
        &:hover {
            background: $panel;
        }
    }
    """


class ToolCall(containers.VerticalGroup):
    DEFAULT_CLASSES = "block"
    DEFAULT_CSS = """
    ToolCall {
        padding: 0 1;        
        width: 1fr;
        layout: stream;
        height: auto;

        .icon {
            width: auto;
            margin-right: 1;
        }
        #tool-content {
            margin-top: 1;            
            display: none;
        }
        &.-expanded {
            #tool-content {
                display: block;
            }
        }
    }

    """

    app = getters.app(ToadApp)
    has_content: var[bool] = var(False)
    expanded: var[bool] = var(False, toggle_class="-expanded")

    def __init__(
        self,
        tool_call: protocol.ToolCall,
        *,
        id: str | None = None,
        classes: str | None = None,
    ) -> None:
        self._tool_call = tool_call
        super().__init__(id=id, classes=classes)

    @property
    def tool_call(self) -> protocol.ToolCall:
        return self._tool_call

    @tool_call.setter
    def tool_call(self, tool_call: protocol.ToolCall):
        self._tool_call = tool_call
        self.refresh(recompose=True)

    def get_block_menu(self) -> Iterable[MenuItem]:
        if self.expanded:
            yield MenuItem("Collapse", "block.collapse", "x")
        else:
            yield MenuItem("Expand", "block.expand", "x")

    def action_collapse(self) -> None:
        self.expanded = False

    def action_expand(self) -> None:
        self.expanded = True

    def get_block_content(self, destination: str) -> str | None:
        return None

    def can_expand(self) -> bool:
        return self.has_content

    def expand_block(self) -> None:
        self.expanded = True

    def collapse_block(self) -> None:
        self.expanded = False

    def is_block_expanded(self) -> bool:
        return self.expanded

    def compose(self) -> ComposeResult:
        tool_call = self._tool_call
        content: list[protocol.ToolCallContent] = tool_call.get("content", None) or []
        self.has_content = bool(content)
        title = tool_call.get("title", "title")

        yield (header := ToolCallHeader(self.tool_call_header_content, markup=False))
        header.tooltip = title
        with containers.VerticalGroup(id="tool-content"):
            yield from self._compose_content(content)

        self.call_after_refresh(self.check_expand)

    def check_expand(self) -> None:
        """Check if the tool call should auto-expand."""
        if not self.has_content:
            return
        tool_call = self._tool_call
        if tool_call.get("kind", "") == "read":
            # Don't auto expand reads, as it can generate a lot of noise
            return
        tool_call_expand = self.app.settings.get("tools.expand", str, expand=False)
        status = self._tool_call.get("status")
        if tool_call_expand == "always":
            self.expanded = True
        elif tool_call_expand != "never" and status is not None:
            if tool_call_expand == "success":
                self.expanded = status == "completed"
            elif tool_call_expand == "fail":
                self.expanded = status == "failed"
            elif tool_call_expand == "both":
                self.expanded = status in ("completed", "failed")

    @property
    def tool_call_header_content(self) -> Content:
        tool_call = self._tool_call
        kind = tool_call.get("kind", "tool")
        title = tool_call.get("title", "title")
        status = tool_call.get("status", "pending")

        expand_icon: Content = Content()
        if self.has_content:
            if self.expanded:
                expand_icon = Content("▼ ")
            else:
                expand_icon = Content("▶ ")
        else:
            expand_icon = Content.styled("▶ ", "dim")

        # header = Content.assemble(
        #     expand_icon,
        #     "🔧 ",
        #     pill(kind, "$primary-muted", "$text-primary"),
        #     " ",
        #     (title, "$text-success"),
        # )

        header = Content.assemble(
            expand_icon,
            "🔧 ",
            (title, "$text-success"),
        )

        if status == "pending":
            header += Content.assemble(" ⏲")
        elif status == "in_progress":
            pass
        elif status == "failed":
            header += Content.assemble(" ", pill("failed", "$error-muted", "$error"))
        elif status == "completed":
            header += Content.from_markup(" [$success]✔")
        return header

    def watch_expanded(self) -> None:
        try:
            self.query_one(ToolCallHeader).update(self.tool_call_header_content)
        except NoMatches:
            pass

    def watch_has_content(self) -> None:
        try:
            self.query_one(ToolCallHeader).update(self.tool_call_header_content)
        except NoMatches:
            pass

    @on(events.Click, "ToolCallHeader")
    def on_click_tool_call_header(self, event: events.Click) -> None:
        event.stop()
        if self.has_content:
            self.expanded = not self.expanded
        else:
            self.app.bell()

    def _compose_content(
        self, tool_call_content: list[protocol.ToolCallContent]
    ) -> ComposeResult:
        def compose_content_block(
            content_block: protocol.ContentBlock,
        ) -> ComposeResult:
            match content_block:
                # TODO: This may need updating
                # Docs claim this should be "plain" text
                # However, I have seen simple text, text with ansi escape sequences, and Markdown returned
                # I think this is a flaw in the spec.
                # For now I will attempt a heuristic to guess what the content actually contains
                # https://agentclientprotocol.com/protocol/schema#param-text
                case {"type": "text", "text": text}:
                    if "\x1b" in text:
                        parsed_ansi_text = Text.from_ansi(text)
                        yield TextContent(Content.from_rich_text(parsed_ansi_text))
                    elif "```" in text or re.search(
                        r"^#{1,6}\s.*$", text, re.MULTILINE
                    ):
                        yield MarkdownContent(text)
                    else:
                        yield TextContent(text, markup=False)

        for content in tool_call_content:
            match content:
                case {"type": "content", "content": sub_content}:
                    yield from compose_content_block(sub_content)
                case {
                    "type": "diff",
                    "path": path,
                    "oldText": old_text,
                    "newText": new_text,
                }:
                    from toad.widgets.diff_view import DiffView

                    yield (diff_view := DiffView(path, path, old_text or "", new_text))

                    if isinstance(self.app, ToadApp):
                        diff_view_setting = self.app.settings.get("diff.view", str)
                        diff_view.split = diff_view_setting == "split"
                        diff_view.auto_split = diff_view_setting == "auto"

                case {"type": "terminal", "terminalId": terminal_id}:
                    pass


if __name__ == "__main__":
    from textual.app import App, ComposeResult

    TOOL_CALL_READ: protocol.ToolCall = {
        "sessionUpdate": "tool_call",
        "toolCallId": "write_file-1759480341499",
        "status": "completed",
        "title": "Foo",
        "content": [
            {
                "type": "diff",
                "path": "fib.py",
                "oldText": "",
                "newText": 'def fibonacci(n):\n    """Generates the Fibonacci sequence up to n terms."""\n    a, b = 0, 1\n    for _ in range(n):\n        yield a\n        a, b = b, a + b\n\nif __name__ == "__main__":\n    for number in fibonacci(10):\n        print(number)\n',
            }
        ],
    }

    TOOL_CALL_CONTENT: protocol.ToolCall = {
        "sessionUpdate": "tool_call",
        "toolCallId": "run_shell_command-1759480356886",
        "status": "completed",
        "title": "Bar",
        "content": [
            {
                "type": "content",
                "content": {
                    "type": "text",
                    "text": "0\n1\n1\n2\n3\n5\n8\n13\n21\n34",
                },
            }
        ],
    }

    TOOL_CALL_EMPTY: protocol.ToolCall = {
        "sessionUpdate": "tool_call",
        "toolCallId": "run_shell_command-1759480356886",
        "status": "completed",
        "title": "Bar",
        "content": [],
    }

    class ToolApp(App):
        def on_mount(self) -> None:
            self.theme = "dracula"

        def compose(self) -> ComposeResult:
            yield ToolCall(TOOL_CALL_READ)
            yield ToolCall(TOOL_CALL_CONTENT)
            yield ToolCall(TOOL_CALL_EMPTY)

    ToolApp().run()



"""Same as the table_movie.py but uses Live to update"""

import time
from contextlib import contextmanager

from rich import box
from rich.align import Align
from rich.console import Console
from rich.live import Live
from rich.table import Table
from rich.text import Text

TABLE_DATA = [
    [
        "May 25, 1977",
        "Star Wars Ep. [b]IV[/]: [i]A New Hope",
        "$11,000,000",
        "$1,554,475",
        "$775,398,007",
    ],
    [
        "May 21, 1980",
        "Star Wars Ep. [b]V[/]: [i]The Empire Strikes Back",
        "$23,000,000",
        "$4,910,483",
        "$547,969,004",
    ],
    [
        "May 25, 1983",
        "Star Wars Ep. [b]VI[/b]: [i]Return of the Jedi",
        "$32,500,000",
        "$23,019,618",
        "$475,106,177",
    ],
    [
        "May 19, 1999",
        "Star Wars Ep. [b]I[/b]: [i]The phantom Menace",
        "$115,000,000",
        "$64,810,870",
        "$1,027,044,677",
    ],
    [
        "May 16, 2002",
        "Star Wars Ep. [b]II[/b]: [i]Attack of the Clones",
        "$115,000,000",
        "$80,027,814",
        "$656,695,615",
    ],
    [
        "May 19, 2005",
        "Star Wars Ep. [b]III[/b]: [i]Revenge of the Sith",
        "$115,500,000",
        "$380,270,577",
        "$848,998,877",
    ],
]

console = Console()

BEAT_TIME = 0.04


@contextmanager
def beat(length: int = 1) -> None:
    yield
    time.sleep(length * BEAT_TIME)


table = Table(show_footer=False)
table_centered = Align.center(table)

console.clear()

with Live(table_centered, console=console, screen=False, refresh_per_second=20):
    with beat(10):
        table.add_column("Release Date", no_wrap=True)

    with beat(10):
        table.add_column("Title", Text.from_markup("[b]Total", justify="right"))

    with beat(10):
        table.add_column("Budget", "[u]$412,000,000", no_wrap=True)

    with beat(10):
        table.add_column("Opening Weekend", "[u]$577,703,455", no_wrap=True)

    with beat(10):
        table.add_column("Box Office", "[u]$4,331,212,357", no_wrap=True)

    with beat(10):
        table.title = "Star Wars Box Office"

    with beat(10):
        table.title = (
            "[not italic]:popcorn:[/] Star Wars Box Office [not italic]:popcorn:[/]"
        )

    with beat(10):
        table.caption = "Made with Rich"

    with beat(10):
        table.caption = "Made with [b]Rich[/b]"

    with beat(10):
        table.caption = "Made with [b magenta not dim]Rich[/]"

    for row in TABLE_DATA:
        with beat(10):
            table.add_row(*row)

    with beat(10):
        table.show_footer = True

    table_width = console.measure(table).maximum

    with beat(10):
        table.columns[2].justify = "right"

    with beat(10):
        table.columns[3].justify = "right"

    with beat(10):
        table.columns[4].justify = "right"

    with beat(10):
        table.columns[2].header_style = "bold red"

    with beat(10):
        table.columns[3].header_style = "bold green"

    with beat(10):
        table.columns[4].header_style = "bold blue"

    with beat(10):
        table.columns[2].style = "red"

    with beat(10):
        table.columns[3].style = "green"

    with beat(10):
        table.columns[4].style = "blue"

    with beat(10):
        table.columns[0].style = "cyan"
        table.columns[0].header_style = "bold cyan"

    with beat(10):
        table.columns[1].style = "magenta"
        table.columns[1].header_style = "bold magenta"

    with beat(10):
        table.columns[2].footer_style = "bright_red"

    with beat(10):
        table.columns[3].footer_style = "bright_green"

    with beat(10):
        table.columns[4].footer_style = "bright_blue"

    with beat(10):
        table.row_styles = ["none", "dim"]

    with beat(10):
        table.border_style = "bright_yellow"

    for box_style in [
        box.SQUARE,
        box.MINIMAL,
        box.SIMPLE,
        box.SIMPLE_HEAD,
    ]:
        with beat(10):
            table.box = box_style

    with beat(10):
        table.pad_edge = False

    original_width = console.measure(table).maximum

    for width in range(original_width, console.width, 2):
        with beat(1):
            table.width = width

    for width in range(console.width, original_width, -2):
        with beat(1):
            table.width = width

    for width in range(original_width, 90, -2):
        with beat(1):
            table.width = width

    for width in range(90, original_width + 1, 2):
        with beat(1):
            table.width = width

    with beat(2):
        table.width = None



from contextlib import suppress
from dataclasses import dataclass
from itertools import zip_longest
import os
from pathlib import Path
from typing import Self

from textual.binding import Binding
from textual.screen import Screen
from textual import events
from textual import work
from textual import getters
from textual import on
from textual.app import ComposeResult
from textual.content import Content
from textual.css.query import NoMatches
from textual.message import Message
from textual import containers
from textual import widgets

import toad
from toad.app import ToadApp
from toad.pill import pill
from toad.widgets.mandelbrot import Mandelbrot
from toad.widgets.grid_select import GridSelect
from toad.agent_schema import Agent
from toad.agents import read_agents


QR = """\
█▀▀▀▀▀█ ▄█ ▄▄█▄▄█ █▀▀▀▀▀█
█ ███ █ ▄█▀█▄▄█▄  █ ███ █
█ ▀▀▀ █ ▄ █ ▀▀▄▄▀ █ ▀▀▀ █
▀▀▀▀▀▀▀ ▀ ▀ ▀ █ █ ▀▀▀▀▀▀▀
█▀██▀ ▀█▀█▀▄▄█   ▀ █ ▀ █ 
 █ ▀▄▄▀▄▄█▄▄█▀██▄▄▄▄ ▀ ▀█
▄▀▄▀▀▄▀ █▀▄▄▄▀▄ ▄▀▀█▀▄▀█▀
█ ▄ ▀▀▀█▀ █ ▀ █▀ ▀ ██▀ ▀█
▀  ▀▀ ▀▀▄▀▄▄▀▀▄▀█▀▀▀█▄▀  
█▀▀▀▀▀█ ▀▄█▄▀▀  █ ▀ █▄▀▀█
█ ███ █ ██▄▄▀▀█▀▀██▀█▄██▄
█ ▀▀▀ █ ██▄▄ ▀  ▄▀ ▄▄█▀ █
▀▀▀▀▀▀▀ ▀▀▀  ▀   ▀▀▀▀▀▀▀▀"""


@dataclass
class LaunchAgent(Message):
    identity: str


class AgentItem(containers.VerticalGroup):
    """An entry in the Agent grid select."""

    def __init__(self, agent: Agent) -> None:
        self._agent = agent
        super().__init__()

    @property
    def agent(self) -> Agent:
        return self._agent

    def compose(self) -> ComposeResult:
        agent = self._agent
        with containers.Grid():
            yield widgets.Label(agent["name"], id="name")
            tag = pill(agent["type"], "$secondary-muted", "$text-primary")
            yield widgets.Label(tag, id="type")
        yield widgets.Label(agent["author_name"], id="author")
        yield widgets.Static(agent["description"], id="description")


class LauncherGridSelect(GridSelect):
    BINDING_GROUP_TITLE = "Launcher"

    app = getters.app(ToadApp)
    BINDINGS = [
        Binding(
            "enter",
            "select",
            "Details",
            tooltip="Open agent details",
        ),
        Binding(
            "space",
            "launch",
            "Lauch",
            tooltip="Launch highlighted agent",
        ),
    ]

    def action_details(self) -> None:
        if self.highlighted is None:
            return
        agent_item = self.children[self.highlighted]
        assert isinstance(agent_item, LauncherItem)
        self.post_message(StoreScreen.OpenAgentDetails(agent_item._agent["identity"]))

    def action_remove(self) -> None:
        agents = self.app.settings.get("launcher.agents", str).splitlines()
        if self.highlighted is None:
            return
        try:
            del agents[self.highlighted]
        except IndexError:
            pass
        else:
            self.app.settings.set("launcher.agents", "\n".join(agents))

    def action_launch(self) -> None:
        if self.highlighted is None:
            return
        child = self.children[self.highlighted]
        assert isinstance(child, LauncherItem)
        self.post_message(LaunchAgent(child.agent["identity"]))


class Launcher(containers.VerticalGroup):
    app = getters.app(ToadApp)
    grid_select = getters.query_one("#launcher-grid-select", GridSelect)
    DIGITS = "123456789ABCDEF"

    def __init__(
        self,
        agents: dict[str, Agent],
        *,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
    ) -> None:
        self._agents = agents
        super().__init__(name=name, id=id, classes=classes)

    @property
    def highlighted(self) -> int | None:
        return self.grid_select.highlighted

    @highlighted.setter
    def highlighted(self, value: int) -> None:
        self.grid_select.highlighted = value

    def focus(self, scroll_visible: bool = True) -> Self:
        try:
            self.grid_select.focus(scroll_visible=scroll_visible)
        except NoMatches:
            pass
        return self

    def compose(self) -> ComposeResult:
        launcher_agents = list(
            dict.fromkeys(
                identity
                for identity in self.app.settings.get(
                    "launcher.agents", str
                ).splitlines()
                if identity.strip()
            )
        )
        agents = self._agents
        self.set_class(not launcher_agents, "-empty")
        with LauncherGridSelect(
            id="launcher-grid-select", min_column_width=32, max_column_width=32
        ):
            for digit, identity in zip_longest(self.DIGITS, launcher_agents):
                if identity is None:
                    break
                yield LauncherItem(digit or "", agents[identity])

        if not launcher_agents:
            yield widgets.Label("Chose your fighter below!", classes="no-agents")


class LauncherItem(containers.VerticalGroup):
    """An entry in the Agent grid select."""

    def __init__(self, digit: str, agent: Agent) -> None:
        self._digit = digit
        self._agent = agent
        super().__init__()

    @property
    def agent(self) -> Agent:
        return self._agent

    def compose(self) -> ComposeResult:
        agent = self._agent
        with containers.HorizontalGroup():
            if self._digit:
                yield widgets.Digits(self._digit)
            with containers.VerticalGroup():
                yield widgets.Label(agent["name"], id="name")
                yield widgets.Label(agent["author_name"], id="author")
                yield widgets.Static(agent["description"], id="description")


class AgentGridSelect(GridSelect):
    BINDINGS = [
        Binding("enter", "select", "Details", tooltip="Open agent details"),
        Binding("space", "launch", "Lauch", tooltip="Launch highlighted agent"),
    ]
    BINDING_GROUP_TITLE = "Agent Select"

    def action_launch(self) -> None:
        if self.highlighted is None:
            return
        child = self.children[self.highlighted]
        assert isinstance(child, AgentItem)
        self.post_message(LaunchAgent(child.agent["identity"]))


class Container(containers.VerticalScroll):
    BINDING_GROUP_TITLE = "View"

    def allow_focus(self) -> bool:
        """Only allow focus when we can scroll."""
        return super().allow_focus() and self.show_vertical_scrollbar


class StoreScreen(Screen):
    BINDING_GROUP_TITLE = "Screen"
    CSS_PATH = "store.tcss"
    FOCUS_GROUP = Binding.Group("Focus")
    BINDINGS = [
        Binding(
            "tab",
            "app.focus_next",
            "Focus Next",
            group=FOCUS_GROUP,
        ),
        Binding(
            "shift+tab",
            "app.focus_previous",
            "Focus Previous",
            group=FOCUS_GROUP,
        ),
        Binding(
            "null",
            "quick_launch",
            "Quick launch",
            key_display="1-9 a-f",
        ),
    ]

    agents_view = getters.query_one("#agents-view", AgentGridSelect)
    launcher = getters.query_one("#launcher", Launcher)

    app = getters.app(ToadApp)

    @dataclass
    class OpenAgentDetails(Message):
        identity: str

    def __init__(
        self, name: str | None = None, id: str | None = None, classes: str | None = None
    ):
        self._agents: dict[str, Agent] = {}
        super().__init__(name=name, id=id, classes=classes)

    @property
    def agents(self) -> dict[str, Agent]:
        return self._agents

    def compose(self) -> ComposeResult:
        with containers.VerticalGroup(id="title-container"):
            with containers.Grid(id="title-grid"):
                yield Mandelbrot()
                yield widgets.Label(self.get_info(), id="info")

        yield Container(id="container")
        yield widgets.Footer()

    def get_info(self) -> Content:
        toad_version = toad.get_version()
        content = Content.assemble(
            Content.from_markup("🐸 Toad"),
            pill(f"v{toad_version}", "$primary-muted", "$text-primary"),
            ("\nThe universal interface for AI in your terminal", "$text-success"),
            (
                "\nSoftware lovingly crafted by hand (with a dash of AI) in Edinburgh, Scotland",
                "dim",
            ),
            "\n",
            (
                Content.from_markup(
                    "\nConsider sponsoring [@click=screen.url('https://github.com/sponsors/willmcgugan')]@willmcgugan[/] to support future updates"
                )
            ),
            "\n\n",
            (
                Content.from_markup(
                    "[dim]Code: [@click=screen.url('https://github.com/Textualize/toad')]Repository[/] "
                    "Bugs: [@click=screen.url('https://github.com/Textualize/toad/discussions')]Discussions[/]"
                )
            ),
        )

        return content

    def action_url(self, url: str) -> None:
        import webbrowser

        webbrowser.open(url)

    def compose_agents(self) -> ComposeResult:
        agents = self._agents

        yield Launcher(agents, id="launcher")

        ordered_agents = sorted(
            agents.values(), key=lambda agent: agent["name"].casefold()
        )

        recommended_agents = [
            agent for agent in ordered_agents if agent.get("recommended", False)
        ]
        if recommended_agents:
            with containers.VerticalGroup(id="sponsored-agents", classes="recommended"):
                yield widgets.Static("Recommended", classes="heading")
                with AgentGridSelect(classes="agents-picker", min_column_width=40):
                    for agent in recommended_agents:
                        yield AgentItem(agent)

        coding_agents = [agent for agent in ordered_agents if agent["type"] == "coding"]
        if coding_agents:
            yield widgets.Static("Coding agents", classes="heading")
            with AgentGridSelect(classes="agents-picker", min_column_width=40):
                for agent in coding_agents:
                    yield AgentItem(agent)

        chat_bots = [agent for agent in ordered_agents if agent["type"] == "chat"]
        if chat_bots:
            yield widgets.Static("Chat & more", classes="heading")
            with AgentGridSelect(classes="agents-picker", min_column_width=40):
                for agent in chat_bots:
                    yield AgentItem(agent)

    @on(GridSelect.Selected, ".agents-picker")
    @work
    async def on_grid_select_selected(self, event: GridSelect.Selected):
        assert isinstance(event.selected_widget, AgentItem)
        from toad.screens.agent_modal import AgentModal

        modal_response = await self.app.push_screen_wait(
            AgentModal(event.selected_widget.agent)
        )
        self.app.save_settings()
        if modal_response == "launch":
            self.post_message(LaunchAgent(event.selected_widget.agent["identity"]))

    @on(OpenAgentDetails)
    @work
    async def open_agent_detail(self, message: OpenAgentDetails) -> None:
        from toad.screens.agent_modal import AgentModal

        try:
            agent = self._agents[message.identity]
        except KeyError:
            return
        modal_response = await self.app.push_screen_wait(AgentModal(agent))
        self.app.save_settings()
        if modal_response == "launch":
            self.post_message(LaunchAgent(agent["identity"]))

    @on(GridSelect.Selected, "#launcher GridSelect")
    @work
    async def on_launcher_selected(self, event: GridSelect.Selected):
        launcher_item = event.selected_widget
        assert isinstance(launcher_item, LauncherItem)

        from toad.screens.agent_modal import AgentModal

        modal_response = await self.app.push_screen_wait(
            AgentModal(launcher_item.agent)
        )
        self.app.save_settings()
        if modal_response == "launch":
            self.post_message(LaunchAgent(launcher_item.agent["identity"]))

    @work
    async def launch_agent(self, agent_identity: str) -> None:
        from toad.screens.main import MainScreen

        agent = self.agents[agent_identity]
        project_path = Path(self.app.project_dir or os.getcwd())
        screen = MainScreen(project_path, agent)
        await self.app.push_screen_wait(screen)

    @on(LaunchAgent)
    def on_launch_agent(self, message: LaunchAgent) -> None:
        self.launch_agent(message.identity)

    @work
    async def on_mount(self) -> None:
        self.app.settings_changed_signal.subscribe(self, self.setting_updated)
        try:
            self._agents = await read_agents()
        except Exception as error:
            self.notify(
                f"Failed to read agents data ({error})",
                title="Agents data",
                severity="error",
            )
        else:
            await self.query_one("#container").mount_compose(self.compose_agents())
            with suppress(NoMatches):
                self.query("GridSelect").first().focus()

    def setting_updated(self, setting: tuple[str, object]) -> None:
        key, value = setting
        if key == "launcher.agents":
            self.launcher.refresh(recompose=True)

    def on_key(self, event: events.Key) -> None:
        if event.character is None:
            return
        LAUNCHER_KEYS = "123456789abcdef"
        if event.character in LAUNCHER_KEYS:
            launch_item_offset = LAUNCHER_KEYS.find(event.character)
            try:
                self.launcher.grid_select.children[launch_item_offset]
            except IndexError:
                self.notify(
                    f"No agent on key [b]{LAUNCHER_KEYS[launch_item_offset]}",
                    title="Quick launch",
                    severity="error",
                )
                self.app.bell()
                return
            self.launcher.focus()
            self.launcher.highlighted = launch_item_offset

    def action_quick_launch(self) -> None:
        self.launcher.focus()


if __name__ == "__main__":
    from toad.app import ToadApp

    app = ToadApp(mode="store")

    app.run()



from collections import defaultdict
from typing import Iterable


class Complete:
    """Stores substrings and their potential completions."""

    def __init__(self) -> None:
        self._word_map: defaultdict[str, set[str]] = defaultdict(set)

    def add_words(self, words: Iterable[str]) -> None:
        """Add word(s) word map.

        Args:
            words: Iterable of words to add.
        """
        word_map = self._word_map
        for word in words:
            for index in range(1, len(word)):
                word_map[word[:index]].add(word[index:])

    def __call__(self, word: str) -> list[str]:
        return sorted(self._word_map.get(word, []), key=len)


if __name__ == "__main__":
    complete = Complete()
    complete.add_words(["ls", "ls -al", "echo 'hello'"])

    print(complete("l"))

    from rich import print

    print(complete._word_map)



import time
import sys

def w(s):
    sys.stdout.write(s)
    sys.stdout.flush()
    time.sleep(0.4)

# Setup: Clear and fill screen
w("\033[2J\033[H")
for i in range(1, 25):
    w(f"Line {i:02d}\n")
time.sleep(1)

# Set scroll region: lines 8-16
w("\033[8;16r")
w("\033[1;1H*** Scroll region: 8-16 ***")
time.sleep(1)

# Position at bottom of scroll region (line 16)
w("\033[16;1H")
w("\033[44m BOTTOM \033[0m")
time.sleep(1)

# IND 3 times - scrolls region UP (line 8 deleted each time)
w("\033[1;1H*** IND x3 (scroll up) ***")
for i in range(3):
    w("\033D")  # IND
    w(f"\033[44m NEW-{i+1} \033[0m")
    time.sleep(0.6)

time.sleep(1)

# Position at top of scroll region (line 8)
w("\033[8;1H")
w("\033[43m TOP \033[0m")
time.sleep(1)

# RI 3 times - scrolls region DOWN (line 16 deleted each time)
w("\033[1;1H*** RI x3 (scroll down) ***")
for i in range(3):
    w("\033M")  # RI
    w(f"\033[43m NEW-{i+1} \033[0m")
    time.sleep(0.6)

# Cleanup
time.sleep(1)
w("\033[r\033[24;1H\n")



import os

print("Hello, World")

def get_terminal_size():
    """Gets the current size of the terminal."""
    try:
        columns, rows = os.get_terminal_size()
        return columns, rows
    except OSError:
        # Return a default size if running in an environment without a TTY
        return 80, 24

def mandelbrot():
    """Generates and prints a Mandelbrot set scaled to the terminal size."""
    WIDTH, HEIGHT = get_terminal_size()
    # Adjust for character aspect ratio (characters are taller than they are wide)
    HEIGHT = int(HEIGHT * 0.5) - 1

    # Region of the complex plane to render
    X_MIN, X_MAX = -2.0, 1.0
    Y_MIN, Y_MAX = -1.0, 1.0
    MAX_ITER = 256

    # Characters used to draw the set, from densest to sparsest
    chars = '@%#*+=-:. '

    for y in range(HEIGHT):
        line = ""
        for x in range(WIDTH):
            # Convert pixel coordinate to a complex number
            c = complex(
                X_MIN + (x / WIDTH) * (X_MAX - X_MIN),
                Y_MIN + (y / HEIGHT) * (Y_MAX - Y_MIN)
            )
            z = 0
            i = 0
            while abs(z) < 2 and i < MAX_ITER:
                z = z*z + c
                i += 1

            if i == MAX_ITER:
                # Point is inside the set
                line += ' '
            else:
                # Point is outside the set; color depends on how quickly it escaped
                line += chars[i % len(chars)]
        print(line)

if __name__ == "__main__":
    mandelbrot()



#!/usr/bin/env python3
"""Simple scroll test - compare with real terminal"""

import sys

ESC = '\x1b'
CSI = f'{ESC}['

def test_scroll_up():
    """Quick scroll up test"""
    # Clear and setup
    print(f'{CSI}2J{CSI}H', end='')  # Clear screen, home cursor
    
    # Draw initial content
    print("BEFORE SCROLL UP:")
    print("Line 1: AAAAA")
    print("Line 2: BBBBB")
    print("Line 3: CCCCC")
    print("Line 4: DDDDD")
    print("Line 5: EEEEE")
    print("\nPress Enter to scroll up 2 lines...")
    input()
    
    # Position to top and scroll
    print(f'{CSI}1;1H', end='')  # Move to row 1, col 1
    print(f'{CSI}2S', end='')     # Scroll up 2 lines
    sys.stdout.flush()
    
    # Show result
    print(f'{CSI}8;1H', end='')  # Move below the content
    print("\nAFTER SCROLL UP 2:")
    print("Line 1 should be: CCCCC")
    print("Line 2 should be: DDDDD")
    print("Line 3 should be: EEEEE")
    print("Line 4 should be: [blank]")
    print("Line 5 should be: [blank]")
    print("\n(Lines AAAAA and BBBBB scrolled off the top)")

def test_scroll_down():
    """Quick scroll down test"""
    input("\n\nPress Enter for scroll down test...")
    
    # Clear and setup
    print(f'{CSI}2J{CSI}H', end='')
    
    # Draw initial content
    print("BEFORE SCROLL DOWN:")
    print("Line 1: AAAAA")
    print("Line 2: BBBBB")
    print("Line 3: CCCCC")
    print("Line 4: DDDDD")
    print("Line 5: EEEEE")
    print("\nPress Enter to scroll down 2 lines...")
    input()
    
    # Position to top and scroll
    print(f'{CSI}1;1H', end='')  # Move to row 1, col 1
    print(f'{CSI}2T', end='')     # Scroll down 2 lines
    sys.stdout.flush()
    
    # Show result
    print(f'{CSI}8;1H', end='')
    print("\nAFTER SCROLL DOWN 2:")
    print("Line 1 should be: [blank]")
    print("Line 2 should be: [blank]")
    print("Line 3 should be: AAAAA")
    print("Line 4 should be: BBBBB")
    print("Line 5 should be: CCCCC")
    print("\n(Lines DDDDD and EEEEE scrolled off the bottom)")

if __name__ == '__main__':
    try:
        test_scroll_up()
        test_scroll_down()
        
        # Cleanup
        print(f'\n{CSI}2J{CSI}H', end='')
        print("Tests complete!")
        
    except KeyboardInterrupt:
        print(f'\n{CSI}2J{CSI}H')
        print("Interrupted")



import os
import sys

import click
from toad.app import ToadApp
from toad.agent_schema import Agent


# Suppress a SyntaxWarning from Posthog
os.environ["PYTHONWARNINGS"] = "ignore::SyntaxWarning"


async def get_agent_data(launch_agent) -> Agent | None:
    launch_agent = launch_agent.lower()

    from toad.agents import read_agents, AgentReadError

    try:
        agents = await read_agents()
    except AgentReadError:
        agents = {}

    for agent_data in agents.values():
        if (
            agent_data["short_name"].lower() == launch_agent
            or agent_data["identity"].lower() == launch_agent
        ):
            launch_agent = agent_data["identity"]
            break

    return agents.get(launch_agent)


class DefaultCommandGroup(click.Group):
    def parse_args(self, ctx, args):
        if "--help" in args or "-h" in args:
            return super().parse_args(ctx, args)
        # Check if first arg is a known subcommand
        if not args or args[0] not in self.commands:
            # If not a subcommand, prepend the default command name
            args.insert(0, "run")
        return super().parse_args(ctx, args)

    def format_usage(self, ctx, formatter):
        formatter.write_usage(ctx.command_path, "[OPTIONS] PATH OR COMMAND [ARGS]...")


@click.group(cls=DefaultCommandGroup)
def main():
    """Toad—AI for your terminal."""


# @click.group(invoke_without_command=True)
# @click.pass_context
@main.command("run")
@click.argument("project_dir", metavar="PATH", required=False, default=".")
@click.option("-a", "--agent", metavar="AGENT", default="")
@click.option(
    "--port",
    metavar="PORT",
    default=8000,
    type=int,
    help="Port to use in conjunction with --serve",
)
@click.option(
    "--host",
    metavar="HOST",
    default="localhost",
    type=str,
    help="Host to use in conjunction with --serve",
)
@click.option("--serve", is_flag=True, help="Serve Toad as a web application")
def run(port: int, host: str, serve: bool, project_dir: str = ".", agent: str = "1"):
    """Run an agent (with also run with `toad PATH`)."""
    # if ctx.invoked_subcommand is not None:
    #     return

    if agent:
        import asyncio

        agent_data = asyncio.run(get_agent_data(agent))
    else:
        agent_data = None

    app = ToadApp(
        mode=None if agent_data else "store",
        agent_data=agent_data,
        project_dir=project_dir,
    )
    if serve:
        import shlex
        from textual_serve.server import Server

        command_args = sys.argv
        try:
            command_args.remove("--serve")
        except ValueError:
            pass
        serve_command = shlex.join(command_args)
        server = Server(
            serve_command,
            host=host,
            port=port,
            title=serve_command,
        )
        server.serve()
    else:
        app.run()
    app.run_on_exit()


@main.command("acp")
@click.argument("command", metavar="COMMAND")
@click.option(
    "--title",
    metavar="TITLE",
    help="Optional title to display in the status bar",
    default=None,
)
@click.option("--project-dir", metavar="PATH", default=None)
@click.option(
    "--port",
    metavar="PORT",
    default=8000,
    type=int,
    help="Port to use in conjunction with --serve",
)
@click.option(
    "--host",
    metavar="HOST",
    default="localhost",
    help="Host to use in conjunction with --serve",
)
@click.option("--serve", is_flag=True, help="Serve Toad as a web application")
def acp(
    command: str,
    host: str,
    port: int,
    title: str | None,
    project_dir: str | None,
    serve: bool = False,
) -> None:
    """Run an ACP client."""

    from rich import print

    from toad.agent_schema import Agent as AgentData

    command_name = command.split(" ", 1)[0].lower()
    identity = f"{command_name}.custom.batrachian.ai"

    agent_data: AgentData = {
        "identity": identity,
        "name": title or command.partition(" ")[0],
        "short_name": "agent",
        "url": "https://github.com/textualize/toad",
        "protocol": "acp",
        "type": "coding",
        "author_name": "Will McGugan",
        "author_url": "https://willmcgugan.github.io/",
        "publisher_name": "Will McGugan",
        "publisher_url": "https://willmcgugan.github.io/",
        "description": "Agent launched from CLI",
        "tags": [],
        "help": "",
        "run_command": {"*": command},
        "actions": {},
    }
    if serve:
        import shlex
        from textual_serve.server import Server

        command_components = [sys.argv[0], "acp", command]
        if project_dir:
            command_components.append(f"--project-dir={project_dir}")
        serve_command = shlex.join(command_components)

        server = Server(
            serve_command,
            host=host,
            port=port,
            title=serve_command,
        )
        server.serve()
    else:
        app = ToadApp(agent_data=agent_data, project_dir=project_dir)
        app.run()
        app.run_on_exit()

    print("")
    print("[bold magenta]Thanks for trying out Toad!")
    print("Please head to Discussions to share your experiences (good or bad).")
    print("https://github.com/Textualize/toad/discussions")


@main.command("settings")
def settings() -> None:
    """Configure settings."""
    app = ToadApp()
    print(f"{app.settings_path}")


# @main.command("replay")
# @click.argument("path", metavar="PATH.jsonl")
# def replay(path: str) -> None:
#     """Replay interaction from a jsonl file."""
#     import time

#     stdout = sys.stdout.buffer
#     with open(path, "rb") as replay_file:
#         for line in replay_file.readlines():
#             time.sleep(0.1)
#             stdout.write(line)
#             stdout.flush()


@main.command("serve")
@click.option("--port", metavar="PORT", default=8000, type=int)
@click.option("--host", metavar="HOST", default="localhost")
def serve(port: int, host: str) -> None:
    """Serve Toad as a web application."""
    from textual_serve.server import Server

    server = Server(sys.argv[0], host=host, port=port, title="Toad")
    server.serve()


@main.command("about")
def about() -> None:
    """Serve Toad as a web application."""

    from toad import about

    app = ToadApp()

    print(about.render(app))


if __name__ == "__main__":
    main()



from typing import Literal

from textual.content import Content
from textual.reactive import var
from textual.widgets import Static
from textual.timer import Timer
from textual import getters


from toad.app import ToadApp


class Flash(Static):
    DEFAULT_CSS = """
    Flash {
        height: 1;
        width: 1fr;
        background: $success 10%;
        color: $text-success;
        text-align: center;
        visibility: hidden;
        text-wrap: nowrap;
        text-overflow: ellipsis;     
        # overlay: screen;
        # offset-y: -1;   

        &.-default {
            background: $primary 10%;
            color: $text-primary;
        }
        
        &.-success {
            background: $success 10%;
            color: $text-success;
        }
        
        
        &.-warning {
            background: $warning 10%;
            color: $text-warning;
        }

        &.-error {
            background: $error 10%;
            color: $text-error;
        }
    }
    """
    app = getters.app(ToadApp)
    flash_timer: var[Timer | None] = var(None)

    def flash(
        self,
        content: str | Content,
        *,
        duration: float | None = None,
        style: Literal["default", "success", "warning", "error"] = "default",
    ) -> None:
        """Flash the content for a brief period.

        Args:
            content: Content to show.
            duration: Duration in seconds to show content.
            style: A semantic style.
        """
        if self.flash_timer is not None:
            self.flash_timer.stop()
        self.visible = False

        def hide() -> None:
            """Hide the content after a while."""
            self.visible = False

        self.update(content)
        self.remove_class("-default", "-success", "-warning", "-error", update=False)
        self.add_class(f"-{style}")
        self.visible = True

        if duration is None:
            duration = self.app.settings.get("ui.flash_duration", float)

        self.flash_timer = self.set_timer(duration or 3, hide)



import time
import sys

def print_slow(s):
    sys.stdout.write(s)
    sys.stdout.flush()
    time.sleep(0.1)

# Clear screen and home cursor
print_slow("\033[2J\033[H")

# Fill screen with numbered lines
for i in range(1, 25):
    sys.stdout.write(f"Line {i-1:02d}\n")
    sys.stdout.flush()

# Set scroll region to lines 5-15
print_slow("\033[5;15r")

# Move cursor to line 15 (bottom of scroll region)
print_slow("\033[15;1H")
print_slow("\033[7m[At bottom of scroll region]\033[0m")
time.sleep(1)

# Use IND (ESC D) - should scroll region UP
for i in range(3):
    print_slow("\033D")  # IND - scrolls up when at bottom
    print_slow(f"\033[7mScroll up {i+1}\033[0m")
    time.sleep(0.2)

time.sleep(1)

# Move cursor to line 5 (top of scroll region)
print_slow("\033[5;1H")
print_slow("\033[7m[At top of scroll region]\033[0m")
time.sleep(1)

# Use RI (ESC M) - should scroll region DOWN
for i in range(3):
    print_slow("\033M")  # RI - scrolls down when at top
    print_slow(f"\033[7mScroll down {i+1}\033[0m")
    time.sleep(0.2)

# Reset scroll region and move to bottom
print_slow("\033[r\033[24;1H")



import sys
import tty
import termios

def get_cursor_position():
    # Save terminal settings
    fd = sys.stdin.fileno()
    old_settings = termios.tcgetattr(fd)
    
    try:
        # Set terminal to raw mode
        tty.setraw(fd)
        
        # Send cursor position request
        sys.stdout.write("\x1B[6n")
        sys.stdout.flush()
        
        # Read response
        response = ""
        while True:
            char = sys.stdin.read(1)
            response += char
            if char == 'R':
                break
        
        # Parse response: ESC[row;colR
        import re
        match = re.search(r'\[(\d+);(\d+)R', response)
        if match:
            row = int(match.group(1))
            col = int(match.group(2))
            return (row, col)
    finally:
        # Restore terminal settings
        termios.tcsetattr(fd, termios.TCSADRAIN, old_settings)


print("\nHello World")
print(get_cursor_position())



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Callable

from textual.app import ComposeResult
from textual import events, on
from textual.binding import Binding
from textual import containers
from textual.content import Content
from textual.reactive import var, reactive
from textual.message import Message
from textual.widget import Widget
from textual.widgets import Label

from toad.answer import Answer

type Options = list[Answer]


@dataclass
class Ask:
    """Data for Question."""

    question: str
    options: Options
    callback: Callable[[Answer], Any] | None = None


class NonSelectableLabel(Label):
    ALLOW_SELECT = False


class Option(containers.HorizontalGroup):
    ALLOW_SELECT = False
    DEFAULT_CSS = """
    Option {

        &:hover {
            background: $boost;
        }
        color: $text-muted;
        #caret {
            visibility: hidden;
            padding: 0 1;
        }
        #index {
            padding-right: 1;
        }
        #label {
            width: 1fr;
        }
        &.-active {            
            color: $text-accent;
            #caret {
                visibility: visible;
            }
        }
        &.-selected {
            opacity: 0.5;
        }
        &.-active.-selected {
            opacity: 1.0;
            background: transparent;
            color: $text-accent;            
            #label {
                text-style: underline;
            }
            #caret {
                visibility: hidden;
            }
        }
    }
    """

    @dataclass
    class Selected(Message):
        """The option was selected."""

        index: int

    selected: reactive[bool] = reactive(False, toggle_class="-selected")

    def __init__(
        self, index: int, content: Content, key: str | None, classes: str = ""
    ) -> None:
        super().__init__(classes=classes)
        self.index = index
        self.content = content
        self.key = key

    def compose(self) -> ComposeResult:
        key = self.key
        yield NonSelectableLabel("❯", id="caret")
        if key:
            yield NonSelectableLabel(Content.styled(f"{key}", "b"), id="index")
        else:
            yield NonSelectableLabel(Content(" "), id="index")

        yield NonSelectableLabel(self.content, id="label")

    def on_click(self, event: events.Click) -> None:
        event.stop()
        self.post_message(self.Selected(self.index))


class Question(Widget, can_focus=True):
    """A text question with a menu of responses."""

    BINDING_GROUP_TITLE = "Question"
    ALLOW_SELECT = False
    CURSOR_GROUP = Binding.Group("Cursor", compact=True)
    ALLOW_GROUP = Binding.Group("Allow once/always", compact=True)
    REJECT_GROUP = Binding.Group("Reject once/always", compact=True)
    BINDINGS = [
        Binding(
            "up",
            "selection_up",
            "Up",
            group=CURSOR_GROUP,
        ),
        Binding(
            "down",
            "selection_down",
            "Down",
            group=CURSOR_GROUP,
        ),
        Binding(
            "enter",
            "select",
            "Select",
        ),
        Binding(
            "a",
            "select_kind(('allow_once', 'allow'))",
            "Allow once",
            group=ALLOW_GROUP,
        ),
        Binding(
            "A",
            "select_kind('allow_always')",
            "Allow always",
            group=ALLOW_GROUP,
        ),
        Binding(
            "r",
            "select_kind(('reject_once', 'reject'))",
            "Reject once",
            group=REJECT_GROUP,
        ),
        Binding(
            "R",
            "select_kind('reject_always')",
            "Reject always",
            group=REJECT_GROUP,
        ),
    ]

    DEFAULT_CSS = """
    Question {
        width: 1fr;
        height: auto;
        padding: 0 1; 
        background: transparent;
        #prompt {
            margin-bottom: 1;
            color: $text-primary;
        }                
        &.-blink Option.-active #caret {
            opacity: 0.2;
        }
        &:blur {
            #index {
                opacity: 0.3;
            }
            #caret {
                opacity: 0.3;
            }
        }
    }
    """

    question: var[str] = var("")
    options: var[Options] = var(list)

    selection: reactive[int] = reactive(0, init=False)
    selected: var[bool] = var(False, toggle_class="-selected")
    blink: var[bool] = var(False)

    DEFAULT_KINDS = {
        "allow_once": "a",
        "allow_always": "A",
        "reject_once": "r",
        "reject_always": "R",
    }

    @dataclass
    class Answer(Message):
        """User selected a response."""

        index: int
        answer: Answer

    def __init__(
        self,
        question: str = "Ask and you will receive",
        options: Options | None = None,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        disabled: bool = False,
    ):
        super().__init__(name=name, id=id, classes=classes, disabled=disabled)
        self.set_reactive(Question.question, question)
        self.set_reactive(Question.options, options or [])

    def on_mount(self) -> None:
        def toggle_blink() -> None:
            if self.has_focus:
                self.blink = not self.blink
            else:
                self.blink = False

        self._blink_timer = self.set_interval(0.5, toggle_blink)

    def _reset_blink(self) -> None:
        self.blink = False
        self._blink_timer.reset()

    def update(self, ask: Ask) -> None:
        self.question = ask.question
        self.options = ask.options
        self.selection = 0
        self.selected = False
        self.refresh(recompose=True, layout=True)

    def compose(self) -> ComposeResult:
        with containers.VerticalGroup():
            if self.question:
                yield Label(self.question, id="prompt")
            with containers.VerticalGroup(id="option-container"):
                for index, answer in enumerate(self.options):
                    active = index == self.selection
                    yield Option(
                        index,
                        Content(answer.text),
                        self.DEFAULT_KINDS.get(str(answer.kind)),
                        classes="-active" if active else "",
                    ).data_bind(Question.selected)

    def watch_selection(self, old_selection: int, new_selection: int) -> None:
        self.query("#option-container > .-active").remove_class("-active")
        if new_selection >= 0:
            self.query_one("#option-container").children[new_selection].add_class(
                "-active"
            )

    def check_action(self, action: str, parameters: tuple[object, ...]) -> bool | None:
        if self.selected and action in ("selection_up", "selection_down"):
            return False
        kinds = {answer.kind for answer in self.options}
        if (
            action in {"allow_once", "allow_always", "reject_once", "reject_always"}
            and parameters[0] not in kinds
        ):
            return False
        return True

    def watch_blink(self, blink: bool) -> None:
        self.set_class(blink, "-blink")

    def action_selection_up(self) -> None:
        self._reset_blink()
        self.selection = max(0, self.selection - 1)

    def action_selection_down(self) -> None:
        self._reset_blink()
        self.selection = min(len(self.options) - 1, self.selection + 1)

    def action_select(self) -> None:
        self._reset_blink()
        self.post_message(
            self.Answer(
                index=self.selection,
                answer=self.options[self.selection],
            )
        )
        self.selected = True

    def action_select_kind(self, kind: str | tuple[str]) -> None:
        kinds = kind if isinstance(kind, tuple) else (kind,)
        for kind in kinds:
            for index, answer in enumerate(self.options):
                print(index, answer)
                if answer.kind == kind:
                    self.selection = index
                    self.action_select()
                    break

    @on(Option.Selected)
    def on_option_selected(self, event: Option.Selected) -> None:
        event.stop()
        self._reset_blink()
        if not self.selected:
            self.selection = event.index


if __name__ == "__main__":
    from textual.app import App
    from textual.widgets import Footer

    OPTIONS = [
        Answer("Yes, allow once", "proceed_always", kind="allow_once"),
        Answer("Yes, allow always", "allow_always", kind="allow_always"),
        Answer("Modify with external editor", "modify"),
        Answer("No, suggest changes (esc)", "reject"),
    ]

    class QuestionApp(App):
        def compose(self) -> ComposeResult:
            yield Question("Apply this change?", OPTIONS)
            yield Footer()

    QuestionApp().run()



#!/usr/bin/env python3
"""
Terminal Scroll Test Script
Tests CSI S (Scroll Up) and CSI T (Scroll Down) commands
"""

import sys
import time

# Escape sequences
ESC = "\x1b"
CSI = f"{ESC}["


def clear_screen():
    """Clear entire screen and move cursor to home"""
    print(f"{CSI}2J", end="")  # Clear screen
    print(f"{CSI}H", end="")  # Move to home (1,1)
    sys.stdout.flush()


def move_cursor(row, col):
    """Move cursor to row, col (1-based)"""
    print(f"{CSI}{row};{col}H", end="")
    sys.stdout.flush()


def set_scroll_region(top, bottom):
    """Set scroll region from top to bottom row"""
    print(f"{CSI}{top};{bottom}r", end="")
    sys.stdout.flush()


def reset_scroll_region():
    """Reset scroll region to full screen"""
    print(f"{CSI}r", end="")
    sys.stdout.flush()


def scroll_up(n=1):
    """Scroll up n lines (CSI n S)"""
    print(f"{CSI}{n}S", end="")
    sys.stdout.flush()


def scroll_down(n=1):
    """Scroll down n lines (CSI n T)"""
    print(f"{CSI}{n}T", end="")
    sys.stdout.flush()


def pause(message="Press Enter to continue..."):
    """Pause and wait for user input"""
    print()
    # input()
    input(message)


def test_basic_scroll_up():
    """Test 1: Basic Scroll Up (Full Screen)"""
    print("=== Test 1: Scroll Up (Full Screen) ===")
    clear_screen()
    move_cursor(1, 1)

    print("Line 1 - Top")
    print("Line 2")
    print("Line 3")
    print("Line 4")
    print("Line 5 - Bottom")

    move_cursor(7, 1)
    print("Before scroll up. Lines 1-5 visible above.")
    pause()

    move_cursor(1, 1)
    scroll_up(2)  # Scroll up 2 lines

    move_cursor(7, 1)
    print("After scrolling up 2 lines:")
    print("  - Lines 1-2 should have scrolled off")
    print("  - Lines 3-5 moved up")
    print("  - 2 blank lines added at bottom")
    pause()


def test_basic_scroll_down():
    """Test 2: Basic Scroll Down (Full Screen)"""
    print("\n=== Test 2: Scroll Down (Full Screen) ===")
    clear_screen()
    move_cursor(1, 1)

    print("Line 1 - Top")
    print("Line 2")
    print("Line 3")
    print("Line 4")
    print("Line 5 - Bottom")

    move_cursor(7, 1)
    print("Before scroll down.")
    pause()

    move_cursor(1, 1)
    scroll_down(2)  # Scroll down 2 lines

    # move_cursor(7, 1)
    # print("After scrolling down 2 lines:")
    # print("  - 2 blank lines inserted at top")
    # print("  - Lines 1-3 moved down")
    # print("  - Lines 4-5 scrolled off bottom")
    # pause()
    input()


def test_scroll_region_up():
    """Test 3: Scroll Up with Scroll Region"""
    print("\n=== Test 3: Scroll Up (With Scroll Region) ===")
    clear_screen()
    move_cursor(1, 1)

    print("Header Line (outside region)")
    print("╔═══ Scroll Region Top ═══╗")
    print("│ Line 1 in region        │")
    print("│ Line 2 in region        │")
    print("│ Line 3 in region        │")
    print("│ Line 4 in region        │")
    print("╚═══ Scroll Region Bottom ═╝")
    print("Footer Line (outside region)")

    move_cursor(10, 1)
    print("Scroll region set to rows 3-7")
    pause()

    set_scroll_region(3, 7)  # Rows 3-7

    move_cursor(10, 1)
    scroll_up(2)

    print("After scroll up 2 in region:")
    print("  - Lines 1-2 scrolled off")
    print("  - Lines 3-4 moved up")
    print("  - Blank lines at bottom of region")
    print("  - Header/Footer unchanged!")

    reset_scroll_region()
    pause()


def test_scroll_region_down():
    """Test 4: Scroll Down with Scroll Region"""
    print("\n=== Test 4: Scroll Down (With Scroll Region) ===")
    clear_screen()
    move_cursor(1, 1)

    print("Header Line (outside region)")
    print("╔═══ Scroll Region Top ═══╗")
    print("│ Line A in region        │")
    print("│ Line B in region        │")
    print("│ Line C in region        │")
    print("│ Line D in region        │")
    print("╚═══ Scroll Region Bottom ═╝")
    print("Footer Line (outside region)")

    move_cursor(10, 1)
    print("Scroll region set to rows 3-7")
    pause()

    set_scroll_region(3, 7)
    move_cursor(10, 1)
    scroll_down(2)

    print("After scroll down 2 in region:")
    print("  - Blank lines at top of region")
    print("  - Lines A-B moved down")
    print("  - Lines C-D scrolled off")
    print("  - Header/Footer unchanged!")

    reset_scroll_region()
    pause()


def test_cursor_preservation():
    """Test 5: Cursor Position Preservation"""
    print("\n=== Test 5: Cursor Position Test ===")
    clear_screen()
    move_cursor(1, 1)

    print("Row 1")
    print("Row 2")
    print("Row 3 <-- Cursor here")
    print("Row 4")
    print("Row 5")

    move_cursor(3, 15)  # Position cursor at row 3, col 15
    print("X", end="")  # Mark cursor position

    move_cursor(7, 1)
    print("Cursor is at row 3, col 16 (after X)")
    pause("Press Enter to scroll up 2 lines...")

    scroll_up(2)

    # print("After scroll up:")
    # print("  - Content scrolled but cursor stayed at row 3, col 16")
    # print("  - (Cursor position doesn't move with scroll)")
    # pause()


def test_sequential_scrolls():
    """Test 6: Multiple Sequential Scrolls"""
    print("\n=== Test 6: Sequential Scrolls ===")
    clear_screen()
    move_cursor(1, 1)

    for i in range(1, 11):
        print(f"Line {i}")

    move_cursor(12, 1)
    print("10 lines displayed")
    pause()

    scroll_up(3)
    move_cursor(12, 1)
    print("Scrolled up 3 lines (lines 1-3 gone)")
    pause()

    scroll_up(3)
    move_cursor(12, 1)
    print("Scrolled up 3 more (lines 4-6 also gone)")
    pause()

    scroll_down(2)
    move_cursor(12, 1)
    print("Scrolled down 2 (blank lines at top)")
    pause()


def test_simple_visual():
    """Simple side-by-side visual test"""
    print("\n=== Simple Visual Test ===")
    clear_screen()

    print("Initial state:")
    print("1: AAAAA")
    print("2: BBBBB")
    print("3: CCCCC")
    print("4: DDDDD")
    print("5: EEEEE")
    print()
    pause("Press Enter, then I'll scroll up 2 lines...")

    # Redraw and scroll
    clear_screen()
    print("1: AAAAA")
    print("2: BBBBB")
    print("3: CCCCC")
    print("4: DDDDD")
    print("5: EEEEE")

    move_cursor(1, 1)
    scroll_up(2)  # Scroll up 2

    move_cursor(7, 1)
    print("After CSI 2 S (scroll up 2):")
    print("Expected result:")
    print("1: CCCCC  (was line 3)")
    print("2: DDDDD  (was line 4)")
    print("3: EEEEE  (was line 5)")
    print("4: [blank]")
    print("5: [blank]")
    pause()


def main():
    """Run all tests"""
    try:
        # Run all tests
        test_basic_scroll_up()
        test_basic_scroll_down()

        test_scroll_region_up()
        test_scroll_region_down()

        test_cursor_preservation()
        test_sequential_scrolls()
        test_simple_visual()

        # Cleanup
        # clear_screen()
        # reset_scroll_region()
        # move_cursor(1, 1)

        # print("Test complete!")
        # print()
        # print("Expected behaviors:")
        # print("  1. Scroll Up (S): Top lines disappear, bottom lines blank")
        # print("  2. Scroll Down (T): Top lines blank, bottom lines disappear")
        # print("  3. Only scroll region contents affected")
        # print("  4. Cursor position preserved")
        # print("  5. Lines outside scroll region unchanged")

    except KeyboardInterrupt:
        print("\n\nTest interrupted by user")
        clear_screen()
        reset_scroll_region()
        move_cursor(1, 1)


if __name__ == "__main__":
    main()



# mypy: disable-error-code="empty-body"
"""
ACP remote API
"""

from toad import jsonrpc
from toad.acp import protocol

API = jsonrpc.API()


@API.method()
def initialize(
    protocolVersion: int,
    clientCapabilities: protocol.ClientCapabilities,
    clientInfo: protocol.Implementation,
) -> protocol.InitializeResponse:
    """https://agentclientprotocol.com/protocol/initialization"""
    ...


@API.method(name="session/new")
def session_new(
    cwd: str, mcpServers: list[protocol.McpServer]
) -> protocol.NewSessionResponse:
    """https://agentclientprotocol.com/protocol/session-setup#session-id"""
    ...


@API.notification(name="session/cancel")
def session_cancel(sessionId: str, _meta: dict):
    """https://agentclientprotocol.com/protocol/prompt-turn#cancellation"""
    ...


@API.method(name="session/prompt")
def session_prompt(
    prompt: list[protocol.ContentBlock], sessionId: str
) -> protocol.SessionPromptResponse:
    """https://agentclientprotocol.com/protocol/prompt-turn#1-user-message"""
    ...


@API.method(name="session/set_mode")
def session_set_mode(sessionId: str, modeId: str) -> protocol.SetSessionModeResponse:
    """https://agentclientprotocol.com/protocol/session-modes#from-the-client"""
    ...



#!/usr/bin/env python3
"""
Terminal Emulator Test Script
Tests various ANSI escape sequences for terminal emulator widgets
"""

import time
import sys


class TerminalTester:
    """Helper class for terminal escape sequences"""
    
    # Cursor movement
    CURSOR_UP = lambda n: f"\033[{n}A"
    CURSOR_DOWN = lambda n: f"\033[{n}B"
    CURSOR_FORWARD = lambda n: f"\033[{n}C"
    CURSOR_BACK = lambda n: f"\033[{n}D"
    CURSOR_POSITION = lambda row, col: f"\033[{row};{col}H"
    CURSOR_SAVE = "\033[s"
    CURSOR_RESTORE = "\033[u"
    CURSOR_HOME = "\033[H"
    
    # Screen operations
    CLEAR_SCREEN = "\033[2J"
    CLEAR_LINE = "\033[2K"
    CLEAR_TO_END = "\033[0J"
    CLEAR_TO_START = "\033[1J"
    
    # Alternate screen
    ALT_SCREEN_ON = "\033[?1049h"
    ALT_SCREEN_OFF = "\033[?1049l"
    
    # Text attributes
    RESET = "\033[0m"
    BOLD = "\033[1m"
    DIM = "\033[2m"
    UNDERLINE = "\033[4m"
    BLINK = "\033[5m"
    REVERSE = "\033[7m"
    
    # Colors
    @staticmethod
    def fg_color(r, g, b):
        return f"\033[38;2;{r};{g};{b}m"
    
    @staticmethod
    def bg_color(r, g, b):
        return f"\033[48;2;{r};{g};{b}m"


def print_slow(text, delay=0.05):
    """Print text character by character with delay"""
    for char in text:
        sys.stdout.write(char)
        sys.stdout.flush()
        time.sleep(delay)


def test_basic_cursor_movement():
    """Test basic cursor movement operations"""
    print("=== Testing Basic Cursor Movement ===\n")
    time.sleep(0.5)
    
    print("Moving cursor forward: ", end="")
    sys.stdout.write(TerminalTester.CURSOR_FORWARD(10))
    print("HERE")
    time.sleep(0.5)
    
    print("Moving cursor back: ")
    sys.stdout.write(TerminalTester.CURSOR_BACK(5))
    print("HERE")
    time.sleep(0.5)
    
    print("\nTesting cursor up/down:")
    print("Line 1")
    print("Line 2")
    print("Line 3")
    sys.stdout.write(TerminalTester.CURSOR_UP(2))
    sys.stdout.write(TerminalTester.CURSOR_FORWARD(7))
    print("<- Moved up 2 lines")
    time.sleep(1)


def test_cursor_positioning():
    """Test absolute cursor positioning"""
    print("\n=== Testing Cursor Positioning ===\n")
    time.sleep(0.5)
    
    # Draw a box using cursor positioning
    positions = [
        (5, 10, "+"), (5, 11, "-"), (5, 12, "-"), (5, 13, "-"), (5, 14, "+"),
        (6, 10, "|"), (6, 14, "|"),
        (7, 10, "|"), (7, 12, "X"), (7, 14, "|"),
        (8, 10, "|"), (8, 14, "|"),
        (9, 10, "+"), (9, 11, "-"), (9, 12, "-"), (9, 13, "-"), (9, 14, "+"),
    ]
    
    for row, col, char in positions:
        sys.stdout.write(TerminalTester.CURSOR_POSITION(row, col))
        sys.stdout.write(char)
        sys.stdout.flush()
        time.sleep(0.05)
    
    # Move cursor below the box
    sys.stdout.write(TerminalTester.CURSOR_POSITION(10, 1))
    print()
    time.sleep(1)


def test_save_restore_cursor():
    """Test save and restore cursor position"""
    print("\n=== Testing Save/Restore Cursor ===\n")
    time.sleep(0.5)
    
    print("Original position", end="")
    sys.stdout.write(TerminalTester.CURSOR_SAVE)
    print(" - cursor saved")
    
    print("Some other text...")
    print("More text on another line...")
    
    sys.stdout.write(TerminalTester.CURSOR_RESTORE)
    print(" <- Restored!")
    time.sleep(1)


def test_screen_clearing():
    """Test screen clearing operations"""
    print("\n=== Testing Screen Clearing ===\n")
    time.sleep(0.5)
    
    for i in range(10):
        print(f"Line {i+1}: " + "x" * 50)
    
    time.sleep(1)
    print("\nClearing screen in 2 seconds...")
    time.sleep(2)
    
    sys.stdout.write(TerminalTester.CLEAR_SCREEN)
    sys.stdout.write(TerminalTester.CURSOR_HOME)
    print("Screen cleared!")
    time.sleep(1)


def test_alternate_screen():
    """Test alternate screen buffer"""
    print("\n=== Testing Alternate Screen ===\n")
    time.sleep(0.5)
    
    print("This is the NORMAL screen buffer")
    print("You should see this text...")
    print("\nSwitching to alternate screen in 2 seconds...")
    time.sleep(2)
    
    # Switch to alternate screen
    sys.stdout.write(TerminalTester.ALT_SCREEN_ON)
    sys.stdout.write(TerminalTester.CLEAR_SCREEN)
    sys.stdout.write(TerminalTester.CURSOR_HOME)
    
    print("╔═══════════════════════════════════╗")
    print("║  ALTERNATE SCREEN BUFFER ACTIVE   ║")
    print("║                                   ║")
    print("║  This is a separate screen!       ║")
    print("║                                   ║")
    print("║  Drawing a simple animation...    ║")
    print("╚═══════════════════════════════════╝")
    
    # Simple animation in alternate screen
    for i in range(5):
        sys.stdout.write(TerminalTester.CURSOR_POSITION(10 + i, 5))
        print(f"{'→' * i}●{'←' * i}")
        sys.stdout.flush()
        time.sleep(0.3)
    
    sys.stdout.write(TerminalTester.CURSOR_POSITION(16, 1))
    print("\nReturning to normal screen in 2 seconds...")
    time.sleep(2)
    
    # Switch back to normal screen
    sys.stdout.write(TerminalTester.ALT_SCREEN_OFF)
    print("Back to normal screen! The previous text should still be here.")
    time.sleep(1)


def test_colors_and_attributes():
    """Test text colors and attributes"""
    print("\n=== Testing Colors and Attributes ===\n")
    time.sleep(0.5)
    
    print(f"{TerminalTester.BOLD}Bold text{TerminalTester.RESET}")
    print(f"{TerminalTester.UNDERLINE}Underlined text{TerminalTester.RESET}")
    print(f"{TerminalTester.REVERSE}Reversed text{TerminalTester.RESET}")
    
    print(f"\n{TerminalTester.fg_color(255, 0, 0)}Red text{TerminalTester.RESET}")
    print(f"{TerminalTester.fg_color(0, 255, 0)}Green text{TerminalTester.RESET}")
    print(f"{TerminalTester.fg_color(0, 0, 255)}Blue text{TerminalTester.RESET}")
    
    print(f"\n{TerminalTester.bg_color(255, 255, 0)}{TerminalTester.fg_color(0, 0, 0)}Black on Yellow{TerminalTester.RESET}")
    
    time.sleep(1)


def test_complex_animation():
    """Test a more complex animation using alternate screen"""
    print("\n=== Complex Animation Test ===\n")
    print("Starting animation in alternate screen in 2 seconds...")
    time.sleep(2)
    
    sys.stdout.write(TerminalTester.ALT_SCREEN_ON)
    sys.stdout.write(TerminalTester.CLEAR_SCREEN)
    
    # Draw border
    width, height = 60, 15
    for col in range(1, width + 1):
        sys.stdout.write(TerminalTester.CURSOR_POSITION(1, col))
        sys.stdout.write("═")
        sys.stdout.write(TerminalTester.CURSOR_POSITION(height, col))
        sys.stdout.write("═")
    
    for row in range(2, height):
        sys.stdout.write(TerminalTester.CURSOR_POSITION(row, 1))
        sys.stdout.write("║")
        sys.stdout.write(TerminalTester.CURSOR_POSITION(row, width))
        sys.stdout.write("║")
    
    sys.stdout.write(TerminalTester.CURSOR_POSITION(1, 1))
    sys.stdout.write("╔")
    sys.stdout.write(TerminalTester.CURSOR_POSITION(1, width))
    sys.stdout.write("╗")
    sys.stdout.write(TerminalTester.CURSOR_POSITION(height, 1))
    sys.stdout.write("╚")
    sys.stdout.write(TerminalTester.CURSOR_POSITION(height, width))
    sys.stdout.write("╝")
    
    # Animate a bouncing ball
    ball_x, ball_y = 30, 7
    dx, dy = 1, 1
    
    for _ in range(50):
        # Clear previous position
        sys.stdout.write(TerminalTester.CURSOR_POSITION(ball_y, ball_x))
        sys.stdout.write(" ")
        
        # Update position
        ball_x += dx
        ball_y += dy
        
        # Bounce off walls
        if ball_x <= 2 or ball_x >= width - 1:
            dx = -dx
        if ball_y <= 2 or ball_y >= height - 1:
            dy = -dy
        
        # Draw ball
        sys.stdout.write(TerminalTester.CURSOR_POSITION(ball_y, ball_x))
        sys.stdout.write(f"{TerminalTester.fg_color(255, 100, 100)}●{TerminalTester.RESET}")
        sys.stdout.flush()
        time.sleep(0.05)
    
    sys.stdout.write(TerminalTester.CURSOR_POSITION(height + 1, 1))
    print("\nAnimation complete! Returning to normal screen...")
    time.sleep(2)
    sys.stdout.write(TerminalTester.ALT_SCREEN_OFF)


def main():
    """Run all tests"""
    print("╔════════════════════════════════════════╗")
    print("║  Terminal Emulator Test Script        ║")
    print("║  Testing various ANSI escape sequences ║")
    print("╚════════════════════════════════════════╝\n")
    time.sleep(1)
    
    try:
        test_basic_cursor_movement()
        test_cursor_positioning()
        test_save_restore_cursor()
        test_colors_and_attributes()
        test_screen_clearing()
        test_alternate_screen()
        test_complex_animation()
        
        print("\n✓ All tests completed!")
        
    except KeyboardInterrupt:
        sys.stdout.write(TerminalTester.ALT_SCREEN_OFF)
        sys.stdout.write(TerminalTester.RESET)
        print("\n\nTests interrupted by user.")
    except Exception as e:
        sys.stdout.write(TerminalTester.ALT_SCREEN_OFF)
        sys.stdout.write(TerminalTester.RESET)
        print(f"\n\nError during tests: {e}")


if __name__ == "__main__":
    main()



def mandelbrot(c, max_iter=100):
    z = 0
    for n in range(max_iter):
        if abs(z) > 2:
            return n
        z = z*z + c
    return max_iter

def generate_mandelbrot(width=80, height=40, x_min=-2.5, x_max=1.5, y_min=-2.0, y_max=2.0):
    chars = " .:-=+*#%@"
    
    for y in range(height):
        line = ""
        for x in range(width):
            # Map pixel coordinates to complex plane
            real = x_min + (x / width) * (x_max - x_min)
            imag = y_min + (y / height) * (y_max - y_min)
            c = complex(real, imag)
            
            # Calculate mandelbrot value
            m = mandelbrot(c)
            
            # Map to character
            char_index = min(int(m / 10), len(chars) - 1)
            line += chars[char_index]
        
        print(line)

if __name__ == "__main__":
    print("Mandelbrot Set")
    print("=" * 80)
    generate_mandelbrot()


from time import sleep

# print("Hello, World", end="", flush=True)
# sleep(1)
# print("\x1b[2K\r", end="", flush=True)
# sleep(1)
# print("Goodbye world")
# sleep(1)

from rich.progress import track
from time import sleep

for n in track(range(10)):
    print(f"Line {n}")
    sleep(0.5)



from __future__ import annotations


import asyncio
import difflib
from itertools import starmap
from typing import Iterable, Literal

from rich.segment import Segment
from rich.style import Style as RichStyle

from textual.app import ComposeResult
from textual.content import Content, Span
from textual.geometry import Size
from textual import highlight
from textual import events

from textual.css.styles import RulesMap
from textual.selection import Selection
from textual.strip import Strip
from textual.style import Style
from textual.reactive import reactive, var
from textual.visual import Visual, RenderOptions
from textual.widget import Widget
from textual.widgets import Static
from textual import containers

type Annotation = Literal["+", "-", "/", " "]


def _format_range_unified(start, stop):
    'Convert range to the "ed" format'
    # Per the diff spec at http://www.unix.org/single_unix_specification/
    beginning = start + 1  # lines start numbering with one
    length = stop - start
    if length == 1:
        return "{}".format(beginning)
    if not length:
        beginning -= 1  # empty ranges begin at line just before the range
    return "{},{}".format(beginning, length)


class DiffScrollContainer(containers.HorizontalGroup):
    scroll_link: var[Widget | None] = var(None)
    DEFAULT_CSS = """
    DiffScrollContainer {
        overflow: scroll hidden;
        scrollbar-size: 0 0;
        height: auto;
    }
    """

    def watch_scroll_x(self, old_value: float, new_value: float) -> None:
        super().watch_scroll_x(old_value, new_value)
        if self.scroll_link:
            self.scroll_link.scroll_x = new_value


class GroupHeading(Static):
    DEFAULT_CSS = """
    GroupHeading {
        width: auto;
        text-opacity: 0.7;
        text-align: center;
        margin-top: 1;
        padding-left: 1;              
    }
    """


class LineContent(Visual):
    def __init__(
        self,
        code_lines: list[Content | None],
        line_styles: list[str],
        width: int | None = None,
    ) -> None:
        self.code_lines = code_lines
        self.line_styles = line_styles
        self._width = width

    def render_strips(
        self, width: int, height: int | None, style: Style, options: RenderOptions
    ) -> list[Strip]:
        strips: list[Strip] = []
        y = 0
        selection = options.selection
        selection_style = options.selection_style or Style.null()
        for y, (line, color) in enumerate(zip(self.code_lines, self.line_styles)):
            if line is None:
                line = Content.styled("╲" * width, "$foreground 15%")
            else:
                if selection is not None:
                    if span := selection.get_span(y):
                        start, end = span
                        if end == -1:
                            end = len(line)
                        line = line.stylize(selection_style, start, end)
                if line.cell_length < width:
                    line = line.pad_right(width - line.cell_length)

            line = line.stylize_before(color).stylize_before(style)
            x = 0
            meta = {"offset": (x, y)}
            segments = []
            for text, rich_style, _ in line.render_segments():
                if rich_style is not None:
                    meta["offset"] = (x, y)
                    segments.append(
                        Segment(text, rich_style + RichStyle.from_meta(meta))
                    )
                else:
                    segments.append(Segment(text, rich_style))
                x += len(text)

            strips.append(Strip(segments, line.cell_length))
        return strips

    def get_optimal_width(self, rules: RulesMap, container_width: int) -> int:
        if self._width is not None:
            return self._width
        return max(line.cell_length for line in self.code_lines if line is not None)

    def get_minimal_width(self, rules: RulesMap) -> int:
        return 1

    def get_height(self, rules: RulesMap, width: int) -> int:
        return len(self.line_styles)


class LineAnnotations(Widget):
    """A vertical strip next to the code, containing line numbers or symbols."""

    DEFAULT_CSS = """
    LineAnnotations {
        width: auto;
        height: auto;                
    }
    """
    numbers: reactive[list[Content]] = reactive(list)

    def __init__(
        self,
        numbers: Iterable[Content],
        *,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        disabled: bool = False,
    ):
        super().__init__(name=name, id=id, classes=classes, disabled=disabled)
        self.numbers = list(numbers)

    @property
    def total_width(self) -> int:
        return self.number_width

    def get_content_width(self, container: Size, viewport: Size) -> int:
        return self.total_width

    def get_content_height(self, container: Size, viewport: Size, width: int) -> int:
        return len(self.numbers)

    @property
    def number_width(self) -> int:
        return max(number.cell_length for number in self.numbers) if self.numbers else 0

    def render_line(self, y: int) -> Strip:
        width = self.total_width
        visual_style = self.visual_style
        rich_style = visual_style.rich_style
        try:
            number = self.numbers[y]
        except IndexError:
            number = Content.empty()

        strip = Strip(
            number.render_segments(visual_style), cell_length=number.cell_length
        )
        strip = strip.adjust_cell_length(width, rich_style)
        return strip


class DiffCode(Static):
    """Container for the code."""

    DEFAULT_CSS = """
    DiffCode {
        width: auto;        
        height: auto;
        min-width: 1fr;
    }
    """
    ALLOW_SELECT = True

    def get_selection(self, selection: Selection) -> tuple[str, str] | None:
        visual = self._render()
        if isinstance(visual, LineContent):
            text = "\n".join(
                "" if line is None else line.plain for line in visual.code_lines
            )
        else:
            return None
        return selection.extract(text), "\n"


def fill_lists[T](list_a: list[T], list_b: list[T], fill_value: T) -> None:
    """Make two lists the same size by extending the smaller with a fill value.

    Args:
        list_a: The first list.
        list_b: The second list.
        fill_value: Value used to extend a list.

    """
    a_length = len(list_a)
    b_length = len(list_b)
    if a_length != b_length:
        if a_length > b_length:
            list_b.extend([fill_value] * (a_length - b_length))
        elif b_length > a_length:
            list_a.extend([fill_value] * (b_length - a_length))


class DiffView(containers.VerticalGroup):
    """A formatted diff in unified or split format."""

    code_before: reactive[str] = reactive("")
    code_after: reactive[str] = reactive("")
    path1: reactive[str] = reactive("")
    path2: reactive[str] = reactive("")
    split: reactive[bool] = reactive(True, recompose=True)
    annotations: var[bool] = var(False, toggle_class="-with-annotations")
    auto_split: var[bool] = var(False)

    DEFAULT_CSS = """
    DiffView {
        width: 1fr;
        height: auto;
        
        .diff-group {
            height: auto;
            background: $foreground 4%;
        }
        
        border-top: wide $foreground 30%;
        border-bottom: wide $foreground 30%;
        border-title-align: center;

        .annotations { width: 1; }
        &.-with-annotations {
            .annotations { width: auto; }
        }
        
    }
    """

    NUMBER_STYLES = {
        "+": "$text-success 80% on $success 20%",
        "-": "$text-error 80% on $error 20%",
        " ": "$foreground 30% on $foreground 3%",
    }
    LINE_STYLES = {
        "+": "on $success 10%",
        "-": "on $error 10%",
        " ": "",
        "/": "",
    }

    def __init__(
        self,
        path1: str,
        path2: str,
        code_before: str,
        code_after: str,
        *,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        disabled: bool = False,
    ):
        super().__init__(name=name, id=id, classes=classes, disabled=disabled)
        self.set_reactive(DiffView.path1, path1)
        self.set_reactive(DiffView.path2, path2)
        self.set_reactive(DiffView.code_before, code_before)
        self.set_reactive(DiffView.code_after, code_after)
        self._grouped_opcodes: list[list[tuple[str, int, int, int, int]]] | None = None
        self._highlighted_code_lines: tuple[list[Content], list[Content]] | None = None

    async def prepare(self) -> None:
        """Do CPU work in a thread.

        Call this method prior to composing or mounting to ensure lazy calculated
        data structures run in a thread. Otherwise the work will be done in the async
        loop, potentially causing a brief freeze.

        """

        def prepare() -> None:
            """Call properties which will lazily update data structures."""
            self.grouped_opcodes
            self.highlighted_code_lines

        await asyncio.to_thread(prepare)

    @property
    def grouped_opcodes(self) -> list[list[tuple[str, int, int, int, int]]]:
        if self._grouped_opcodes is None:
            text_lines_a = self.code_before.splitlines()
            text_lines_b = self.code_after.splitlines()
            sequence_matcher = difflib.SequenceMatcher(
                lambda character: character in " \t",
                text_lines_a,
                text_lines_b,
                autojunk=True,
            )
            self._grouped_opcodes = list(sequence_matcher.get_grouped_opcodes())

        return self._grouped_opcodes

    @property
    def counts(self) -> tuple[int, int]:
        """Additions and removals."""
        additions = 0
        removals = 0
        for group in self.grouped_opcodes:
            for tag, i1, i2, j1, j2 in group:
                if tag == "delete":
                    removals += 1
                elif tag == "replace":
                    additions += 1
                    removals += 1
                elif tag == "insert":
                    additions += 1
        return additions, removals

    @property
    def highlighted_code_lines(self) -> tuple[list[Content], list[Content]]:
        """Get syntax highlighted code for both files, as a list of lines.

        Returns:
            A pair of line lists for `code_before` and `code_after`
        """
        if self._highlighted_code_lines is None:
            language1 = highlight.guess_language(self.code_before, self.path1)
            language2 = highlight.guess_language(self.code_after, self.path2)
            text_lines_a = self.code_before.splitlines()
            text_lines_b = self.code_after.splitlines()

            code_a = highlight.highlight(
                "\n".join(text_lines_a), language=language1, path=self.path1
            )
            code_b = highlight.highlight(
                "\n".join(text_lines_b), language=language2, path=self.path2
            )

            sequence_matcher = difflib.SequenceMatcher(
                lambda character: character in " \t",
                self.code_before,
                self.code_after,
                autojunk=True,
            )
            code_a_spans: list[Span] = []
            code_b_spans: list[Span] = []
            for tag, i1, i2, j1, j2 in sequence_matcher.get_opcodes():
                if tag == "delete" and "\n" not in code_a.plain[i1 : i2 + 1]:
                    code_a_spans.append(Span(i1, i2, "on $error 40%"))

                if tag == "insert" and "\n" not in code_b.plain[j1 : j2 + 1]:
                    code_b_spans.append(Span(j1, j2, "on $success 40%"))

            code_a = code_a.add_spans(code_a_spans)
            code_b = code_b.add_spans(code_b_spans)

            lines_a = code_a.split("\n")
            lines_b = code_b.split("\n")
            self._highlighted_code_lines = (lines_a, lines_b)
        return self._highlighted_code_lines

    def get_title(self) -> Content:
        """Get a title for the diff view.

        Returns:
            A Content instance.
        """
        additions, removals = self.counts
        title = Content.from_markup(
            "[b]$path[/b] [b]([/b][$text-success][b]$additions[/b] $additions_label[/], [$text-error][b]$removals[/b] $removals_label[/][b])[/b]",
            path=self.path2,
            additions=additions,
            removals=removals,
            additions_label="addition" if additions == 1 else "additions",
            removals_label="removals" if removals == 1 else "removals",
        ).stylize_before("$text")
        return title

    def compose(self) -> ComposeResult:
        """Compose either split or unified view."""

        self.border_title = self.get_title()
        # yield DiffTitle(self.path2, additions, removals)
        if self.split:
            yield from self.compose_split()
        else:
            yield from self.compose_unified()

    def _check_auto_split(self, width: int):
        if self.auto_split:
            lines_a, lines_b = self.highlighted_code_lines
            split_width = max([line.cell_length for line in (lines_a + lines_b)]) * 2
            split_width += 4 + 2 * (
                max(
                    [
                        len(str(len(lines_a))),
                        len(str(len(lines_b))),
                    ]
                )
            )
            split_width += 3 * 2 if self.annotations else 2
            self.split = width >= split_width

    async def on_resize(self, event: events.Resize) -> None:
        self._check_auto_split(event.size.width)

    async def on_mount(self) -> None:
        self._check_auto_split(self.size.width)

    def compose_unified(self) -> ComposeResult:
        lines_a, lines_b = self.highlighted_code_lines

        for group in self.grouped_opcodes:
            line_numbers_a: list[int | None] = []
            line_numbers_b: list[int | None] = []
            annotations: list[str] = []
            code_lines: list[Content | None] = []
            first, last = group[0], group[-1]
            file1_range = _format_range_unified(first[1], last[2])
            file2_range = _format_range_unified(first[3], last[4])
            yield GroupHeading(
                "@@ [$text-error]-{}[/] [$text-success]+{}[/] @@".format(
                    file1_range, file2_range
                )
            )
            for tag, i1, i2, j1, j2 in group:
                if tag == "equal":
                    for line_offset, line in enumerate(lines_a[i1:i2], 1):
                        annotations.append(" ")
                        line_numbers_a.append(i1 + line_offset)
                        line_numbers_b.append(j1 + line_offset)
                        code_lines.append(line)
                    continue
                if tag in {"replace", "delete"}:
                    for line_offset, line in enumerate(lines_a[i1:i2], 1):
                        annotations.append("-")
                        line_numbers_a.append(i1 + line_offset)
                        line_numbers_b.append(None)
                        code_lines.append(line)
                if tag in {"replace", "insert"}:
                    for line_offset, line in enumerate(lines_b[j1:j2], 1):
                        annotations.append("+")
                        line_numbers_a.append(None)
                        line_numbers_b.append(j1 + line_offset)
                        code_lines.append(line)

            NUMBER_STYLES = self.NUMBER_STYLES
            LINE_STYLES = self.LINE_STYLES

            line_number_width = max(
                len("" if line_no is None else str(line_no))
                for line_no in (line_numbers_a + line_numbers_b)
            )

            with containers.HorizontalGroup(classes="diff-group"):
                yield LineAnnotations(
                    [
                        (
                            Content(f" {' ' * line_number_width} ")
                            if line_no is None
                            else Content(f" {line_no:>{line_number_width}} ")
                        ).stylize(NUMBER_STYLES[annotation])
                        for line_no, annotation in zip(line_numbers_a, annotations)
                    ]
                )

                yield LineAnnotations(
                    [
                        (
                            Content(f" {' ' * line_number_width} ")
                            if line_no is None
                            else Content(f" {line_no:>{line_number_width}} ")
                        ).stylize(NUMBER_STYLES[annotation])
                        for line_no, annotation in zip(line_numbers_b, annotations)
                    ]
                )

                yield LineAnnotations(
                    [
                        (Content(f" {annotation} "))
                        .stylize(LINE_STYLES[annotation])
                        .stylize("bold")
                        for annotation in annotations
                    ],
                    classes="annotations",
                )
                code_line_styles = [
                    LINE_STYLES[annotation] for annotation in annotations
                ]
                with DiffScrollContainer():
                    yield DiffCode(LineContent(code_lines, code_line_styles))

    def compose_split(self) -> ComposeResult:
        lines_a, lines_b = self.highlighted_code_lines

        annotation_hatch = Content.styled("╲" * 3, "$foreground 15%")
        annotation_blank = Content(" " * 3)

        def make_annotation(
            annotation: Annotation, highlight_annotation: Literal["+", "-"]
        ) -> Content:
            """Format an annotation.

            Args:
                annotation: Annotation to format.
                highlight_annotation: Annotation to highlight ('+' or '-')

            Returns:
                Content with annotation.
            """
            if annotation == highlight_annotation:
                return (
                    Content(f" {annotation} ")
                    .stylize(self.LINE_STYLES[annotation])
                    .stylize("bold")
                )
            if annotation == "/":
                return annotation_hatch
            return annotation_blank

        for group in self.grouped_opcodes:
            first, last = group[0], group[-1]
            file1_range = _format_range_unified(first[1], last[2])
            file2_range = _format_range_unified(first[3], last[4])
            yield GroupHeading(
                "@@ [$text-error]-{}[/] [$text-success]+{}[/] @@".format(
                    file1_range, file2_range
                )
            )
            line_numbers_a: list[int | None] = []
            line_numbers_b: list[int | None] = []
            annotations_a: list[Annotation] = []
            annotations_b: list[Annotation] = []
            code_lines_a: list[Content | None] = []
            code_lines_b: list[Content | None] = []
            for tag, i1, i2, j1, j2 in group:
                if tag == "equal":
                    for line_offset, line in enumerate(lines_a[i1:i2], 1):
                        annotations_a.append(" ")
                        annotations_b.append(" ")
                        line_numbers_a.append(i1 + line_offset)
                        line_numbers_b.append(j1 + line_offset)
                        code_lines_a.append(line)
                        code_lines_b.append(line)
                else:
                    if tag in {"replace", "delete"}:
                        for line_number, line in enumerate(lines_a[i1:i2], i1 + 1):
                            annotations_a.append("-")
                            line_numbers_a.append(line_number)
                            code_lines_a.append(line)
                    if tag in {"replace", "insert"}:
                        for line_number, line in enumerate(lines_b[j1:j2], j1 + 1):
                            annotations_b.append("+")
                            line_numbers_b.append(line_number)
                            code_lines_b.append(line)
                    fill_lists(code_lines_a, code_lines_b, None)
                    fill_lists(annotations_a, annotations_b, "/")
                    fill_lists(line_numbers_a, line_numbers_b, None)

            if line_numbers_a or line_numbers_b:
                line_number_width = max(
                    0 if line_no is None else len(str(line_no))
                    for line_no in (line_numbers_a + line_numbers_b)
                )
            else:
                line_number_width = 1

            hatch = Content.styled("╲" * (2 + line_number_width), "$foreground 15%")

            def format_number(line_no: int | None, annotation: str) -> Content:
                """Format a line number with an annotation.

                Args:
                    line_no: Line number or `None` if there is no line here.
                    annotation: An annotation string ('+', '-', or ' ')

                Returns:
                    Content for use in the `LineAnnotations` widget.
                """
                return (
                    hatch
                    if line_no is None
                    else Content(f" {line_no:>{line_number_width}} ").stylize(
                        self.NUMBER_STYLES[annotation]
                    )
                )

            with containers.HorizontalGroup(classes="diff-group"):
                # Before line numbers
                yield LineAnnotations(
                    starmap(format_number, zip(line_numbers_a, annotations_a))
                )
                # Before annotations
                yield LineAnnotations(
                    [make_annotation(annotation, "-") for annotation in annotations_a],
                    classes="annotations",
                )

                code_line_styles = [
                    self.LINE_STYLES[annotation] for annotation in annotations_a
                ]
                line_width = max(
                    line.cell_length
                    for line in code_lines_a + code_lines_b
                    if line is not None
                )
                # Before code
                with DiffScrollContainer() as scroll_container_a:
                    yield DiffCode(
                        LineContent(code_lines_a, code_line_styles, width=line_width)
                    )

                # After line numbers
                yield LineAnnotations(
                    starmap(format_number, zip(line_numbers_b, annotations_b))
                )
                # After annotations
                yield LineAnnotations(
                    [make_annotation(annotation, "+") for annotation in annotations_b],
                    classes="annotations",
                )

                code_line_styles = [
                    self.LINE_STYLES[annotation] for annotation in annotations_b
                ]
                # After code
                with DiffScrollContainer() as scroll_container_b:
                    yield DiffCode(
                        LineContent(code_lines_b, code_line_styles, width=line_width)
                    )

                # Link scroll containers, so they scroll together
                scroll_container_a.scroll_link = scroll_container_b
                scroll_container_b.scroll_link = scroll_container_a


if __name__ == "__main__":
    SOURCE1 = '''\
def loop_first(values: Iterable[T]) -> Iterable[tuple[bool, T]]:
    """Iterate and generate a tuple with a flag for first value."""
    iter_values = iter(values)
    try:
        value = next(iter_values)
    except StopIteration:
        return
    yield True, value
    for value in iter_values:
        yield False, value


def loop_first_last(values: Iterable[T]) -> Iterable[tuple[bool, T]]:
    """Iterate and generate a tuple with a flag for first and last value."""
    iter_values = iter(values)
    try:
        previous_value = next(iter_values)
    except StopIteration:
        return
    first = True
    for value in iter_values:
        yield first, False, previous_value
        first = False
        previous_value = value
    yield first, True, previous_value

'''

    SOURCE2 = '''\
def loop_first(values: Iterable[T]) -> Iterable[tuple[bool, T]]:
    """Iterate and generate a tuple with a flag for first value.
    
    Args:
        values: iterables of values.

    Returns:
        Iterable of a boolean to indicate first value, and a value from the iterable.
    """
    iter_values = iter(values)
    try:
        value = next(iter_values)
    except StopIteration:
        return
    yield True, value
    for value in iter_values:
        yield False, value


def loop_last(values: Iterable[T]) -> Iterable[tuple[bool, bool, T]]:
    """Iterate and generate a tuple with a flag for last value."""
    iter_values = iter(values)
    try:
        previous_value = next(iter_values)
    except StopIteration:
        return
    for value in iter_values:
        yield False, previous_value
        previous_value = value
    yield True, previous_value


def loop_first_last(values: Iterable[ValueType]) -> Iterable[tuple[bool, bool, ValueType]]:
    """Iterate and generate a tuple with a flag for first and last value."""
    iter_values = iter(values)
    try:
        previous_value = next(iter_values)  # Get previous value
    except StopIteration:
        return
    first = True

'''
    from textual.app import App
    from textual.widgets import Footer

    class DiffApp(App):
        BINDINGS = [
            ("space", "split", "Toggle split"),
            ("a", "toggle_annotations", "Toggle annotations"),
        ]

        def compose(self) -> ComposeResult:
            yield DiffView("foo.py", "foo.py", SOURCE1, SOURCE2)
            yield Footer()

        def action_split(self) -> None:
            self.query_one(DiffView).split = not self.query_one(DiffView).split

        def action_toggle_annotations(self) -> None:
            self.query_one(DiffView).annotations = not self.query_one(
                DiffView
            ).annotations

    app = DiffApp()
    app.run()



from __future__ import annotations

import asyncio
import fnmatch
from typing import Callable, Iterable, Sequence
from time import time
from os import PathLike
from pathlib import Path

from pathspec import PathSpec


class ScanJob:
    """A single directory scanning job."""

    def __init__(
        self,
        name: str,
        queue: asyncio.Queue[Path],
        results: list[Path],
        exclude_dirs: Sequence[str],
        exclude_files: Sequence[str],
        path_spec: PathSpec | None = None,
        add_directories=False,
    ) -> None:
        self.queue = queue
        self.results = results
        self.exclude_dirs = exclude_dirs
        self.exclude_files = exclude_files
        self.name = name
        self.path_spec = path_spec
        self.add_directories = add_directories

    def start(self) -> None:
        self._task = asyncio.create_task(self.run())

    async def is_file(self, path: Path) -> bool:
        """Check if the path references a file.

        Args:
            path: A path.

        Returns:
            `True` if the path is a file, `False` if it isn't or an error occured.
        """
        try:
            return await asyncio.to_thread(path.is_file)
        except OSError:
            return False

    async def is_dir(self, path: Path) -> bool:
        """Check if the path references a directory.

        Args:
            path: A path.

        Returns:
            `True` if the path is a directory, `False` if it isn't or an error occured.
        """
        try:
            return await asyncio.to_thread(path.is_dir)
        except OSError:
            return False

    async def run(self) -> None:
        queue = self.queue
        results = self.results
        add_directories = self.add_directories
        while True:
            try:
                scan_path = await queue.get()
            except asyncio.QueueShutDown:
                break
            paths = await self._scan(scan_path)
            for path in paths:
                if self.path_spec is not None and self.path_spec.match_file(path):
                    continue
                if await self.is_file(path):
                    str_path = str(path.name)
                    for exclude in self.exclude_files:
                        if fnmatch.fnmatch(str_path, exclude):
                            break
                    else:
                        results.append(path)
                elif await self.is_dir(path):
                    str_path = str(path.name)
                    for exclude in self.exclude_dirs:
                        if fnmatch.fnmatch(str_path, exclude):
                            break
                    else:
                        if add_directories:
                            results.append(path)
                        await queue.put(path)
            queue.task_done()

    async def _scan(self, root: Path) -> list[Path]:
        """Get a directory listing.

        Args:
            root: Root path.

        Returns:
            List of paths within the given directory, or empty list if an error occured.
        """

        def get_directory() -> list[Path]:
            try:
                return list(root.iterdir())
            except IOError:
                return []

        return await asyncio.to_thread(get_directory)


async def scan(
    root: Path,
    *,
    max_simultaneous: int = 5,
    exclude_dirs: Sequence[str] | None = None,
    exclude_files: Sequence[str] | None = None,
    path_spec: PathSpec | None = None,
    add_directories: bool = False,
) -> list[Path]:
    """Scan a directory for paths.

    Args:
        root: Root directory to scan.
        max_simultaneous: Maximum number of scan jobs.
        exclude_dirs: Wildcards to exclude directories.
        exclude_files: Wildcards to exclude paths.

    Returns:
        A list of Paths.
    """
    queue: asyncio.Queue[Path] = asyncio.Queue()
    results: list[Path] = []
    jobs = [
        ScanJob(
            f"scan-job #{index}",
            queue,
            results,
            exclude_dirs=exclude_dirs or [],
            exclude_files=exclude_files or [],
            path_spec=path_spec,
            add_directories=add_directories,
        )
        for index in range(max_simultaneous)
    ]
    try:
        await queue.put(root)
        for job in jobs:
            job.start()
        await queue.join()
    except asyncio.CancelledError:
        await queue.join()
    queue.shutdown(immediate=True)
    return results


class Scan:
    """A scan of a single directory."""

    def __init__(self, root: Path, on_complete: Callable[[Scan]]) -> None:
        self.root = root
        self._on_complete = on_complete
        self._complete_event = asyncio.Event()
        self._scan_result: list[Path] = []
        self._scan_task: asyncio.Task | None = None
        self._scan_time = time()

    @property
    def is_complete(self) -> bool:
        """Has the scan finished?"""
        return self._complete_event.is_set()

    def start(self) -> None:
        self._scan_time = time()
        self._scan_task = asyncio.create_task(self._run(), name=f"scan {self.root!s}")

    async def _run(self) -> None:
        await asyncio.to_thread(self._scan)
        self._complete_event.set()
        self._on_complete(self)

    def _scan(self) -> None:
        self._scan_result = list(self.root.iterdir())

    async def wait(self) -> list[Path]:
        """Get the result of the scan, potentially waiting for it to finish first.

        Returns:
            A list of paths in the root.
        """
        await self._complete_event.wait()
        assert self._scan_result is not None
        self._scan_task = None
        return self._scan_result


class DirectoryScanner:
    def __init__(self, root: PathLike) -> None:
        self.root = Path(root)
        self.directories: dict[Path, Scan] = {}

    async def scan(
        self, relative_directory_path: str, on_complete: Callable[[Scan]]
    ) -> Scan:
        """Get a scan.

        Scans are created on demand, or returned previously scanned.

        Args:
            relative_directory_path: A path relative to the root.
            on_complete: Callback when scan is complete, will be invoked with Scan instance.

        Returns:
            A scan instance.
        """
        scan_path = self.root / relative_directory_path
        if scan := self.directories.get(scan_path):
            if scan.is_complete:
                on_complete(scan)
        else:
            self.directories[scan_path] = scan = Scan(
                scan_path, on_complete=on_complete
            )
            scan.start()
        return scan


if __name__ == "__main__":
    # from rich import print

    from textual.fuzzy import Matcher

    import contextlib
    from time import perf_counter
    from typing import Generator

    @contextlib.contextmanager
    def timer(subject: str = "time") -> Generator[None, None, None]:
        """print the elapsed time. (only used in debugging)"""
        start = perf_counter()
        yield
        elapsed = perf_counter() - start
        elapsed_ms = elapsed * 1000
        print(f"{subject} elapsed {elapsed_ms:.4f}ms")

    async def run_scan():
        paths = await scan(
            Path("./"),
            # exclude_dirs=[".*", "__pycache__"],
        )
        str_paths = [str(path) for path in paths]
        matcher = Matcher("psputils")
        results = []

        with timer("fuzzy"):
            for path in str_paths:
                score = matcher.match(path)
                if score > 0:
                    print(path)
                results.append((score, path))

        # print(results)

    asyncio.run(run_scan())
